from enum import Enum
from typing import TYPE_CHECKING, Annotated, Any, Literal

from pydantic import ConfigDict, Discriminator, Tag

from pbi_core.lineage import LineageNode
from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.filters import Filter, PrototypeQuery
from pbi_core.static_files.layout.selector import Selector
from pbi_core.static_files.layout.sources.column import ColumnSource
from pbi_core.static_files.layout.sources.measure import MeasureSource
from pbi_core.static_files.layout.sources.paragraphs import Paragraph
from pbi_core.static_files.layout.visuals.properties import Expression

from .properties.vc_properties import VCProperties

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables import Column, Measure
    from pbi_core.ssas.server import BaseTabularModel
    from pbi_core.static_files.model_references import ModelColumnReference, ModelMeasureReference


class FilterSortOrder(Enum):
    NA = 0
    NA1 = 1
    NA2 = 2
    NA3 = 3


class DisplayMode(Enum):
    hidden = "hidden"


class Display(LayoutNode):
    mode: DisplayMode


class ProjectionConfig(LayoutNode):
    queryRef: str
    active: bool = False
    suppressConcat: bool = False


class ColorRule1(LayoutNode):
    positiveColor: Expression
    negativeColor: Expression
    axisColor: Expression
    reverseDirection: Expression
    hideText: Expression | None = None


def get_property_expression(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "positiveColor" in v:
            return "ColorRule1"
        if "filter" in v:
            return "Filter"
        return "Expression"
    if isinstance(v, list):
        return "Paragraph"
    return v.__class__.__name__


PropertyExpression = Annotated[
    Annotated[Expression, Tag("Expression")]
    | Annotated[Filter, Tag("Filter")]
    | Annotated[ColorRule1, Tag("ColorRule1")]
    | Annotated[list[Paragraph], Tag("Paragraph")],
    Discriminator(get_property_expression),
]


class PropertyDef(LayoutNode):
    properties: dict[str, PropertyExpression]
    selector: Selector | None = None


class QueryOptions(LayoutNode):
    allowOverlappingPointsSample: bool = False
    keepProjectionOrder: bool = True


class BaseVisual(LayoutNode):
    """Base class for all visual representations in the layout.

    This class defines the common properties and methods that all visuals should implement. It
    serves as a foundation for more specific visual types.

    Note:
        This class is not intended to be instantiated directly. Instead, it should be subclassed

    """

    model_config = ConfigDict(extra="allow")

    prototypeQuery: PrototypeQuery | None = None
    projections: dict[str, list[ProjectionConfig]] | None = None
    hasDefaultSort: bool = False
    drillFilterOtherVisuals: bool = False
    filterSortOrder: FilterSortOrder = FilterSortOrder.NA
    # vcObjects means "visual container objects"
    vcObjects: VCProperties | None = None
    visualType: str = "unknown"
    queryOptions: QueryOptions | None = None
    showAllRoles: list[str] | None = None
    display: Display | None = None

    @property
    def id(self) -> str:
        """Obviously terrible, but works for now lol."""
        return self.visualType

    def pbi_core_name(self) -> str:
        """Returns the name displayed in the PBIX report."""
        return self.__class__.__name__

    def get_ssas_elements(self) -> "set[ModelColumnReference | ModelMeasureReference]":
        if self.prototypeQuery is None:
            return set()
        return self.prototypeQuery.get_ssas_elements()

    def get_lineage(
        self,
        lineage_type: Literal["children", "parents"],
        tabular_model: "BaseTabularModel",
    ) -> LineageNode:
        ret: list[Column | Measure] = []
        if self.prototypeQuery is None:
            return LineageNode(self, lineage_type, [])
        table_mapping = self.prototypeQuery.table_mapping()
        children = self.prototypeQuery.get_ssas_elements()
        for child in children:
            if isinstance(child, ColumnSource):
                candidate_columns = tabular_model.columns.find_all({"explicit_name": child.Column.column()})
                for candidate_column in candidate_columns:
                    if candidate_column.table().name == table_mapping[child.Column.table()]:
                        ret.append(candidate_column)
                        break
            elif isinstance(child, MeasureSource):
                candidate_measures = tabular_model.measures.find_all({"name": child.Measure.column()})
                for candidate_measure in candidate_measures:
                    if candidate_measure.table().name == table_mapping[child.Measure.table()]:
                        ret.append(candidate_measure)
                        break
        return LineageNode(self, lineage_type, relatives=[child.get_lineage(lineage_type) for child in ret])



from collections.abc import Callable
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal, TypeVar

from pbi_core.lineage import LineageNode
from pbi_core.pydantic.main import BaseValidation

if TYPE_CHECKING:
    from pbi_core.ssas.server import BaseTabularModel

LAYOUT_ENCODING = "utf-16-le"


T = TypeVar("T", bound="LayoutNode")


class LayoutNode(BaseValidation):
    _name_field: str | None = None  # name of the field used to populate __repr__
    _parent: "LayoutNode"
    _xpath: list[str | int]

    @staticmethod
    def serialize_helper(value: Any) -> Any:
        """Helper function to serialize a value.

        We need to separate from the main function to handle cases where there is a list of
        dictionaries such as the visual container properties.
        """
        if hasattr(value, "serialize"):
            return value.serialize()
        if isinstance(value, list):
            return [LayoutNode.serialize_helper(val) for val in value]
        if isinstance(value, dict):
            return {key: LayoutNode.serialize_helper(val) for key, val in value.items()}
        if isinstance(value, Enum):
            return value.name
        return value

    def serialize(self) -> dict[str, Any]:
        """Serialize the node to a dictionary.

        Differs from the model_dump_json method in that it does not convert the JSON models back to strings.
        """
        ret = {}
        for field in self.model_fields:
            ret[field] = self.serialize_helper(getattr(self, field))
        return ret

    def pbi_core_name(self) -> str:
        raise NotImplementedError

    def find_all(
        self,
        cls_type: type[T],
        attributes: dict[str, Any] | Callable[[T], bool] | None = None,
    ) -> list["T"]:
        ret: list[T] = []
        if attributes is None:
            attribute_lambda: Callable[[T], bool] = lambda _: True  # noqa: E731
        elif isinstance(attributes, dict):
            attribute_lambda = lambda x: all(  # noqa: E731
                getattr(x, field_name) == field_value for field_name, field_value in attributes.items()
            )
        else:
            attribute_lambda = attributes
        if isinstance(self, cls_type) and attribute_lambda(self):
            ret.append(self)
        for child in self._children():
            ret.extend(child.find_all(cls_type, attributes))
        return ret

    def find(self, cls_type: type[T], attributes: dict[str, Any] | Callable[[T], bool] | None = None) -> "T":
        if attributes is None:
            attribute_lambda: Callable[[T], bool] = lambda _: True  # noqa: E731
        elif isinstance(attributes, dict):
            attribute_lambda = lambda x: all(  # noqa: E731
                getattr(x, field_name) == field_value for field_name, field_value in attributes.items()
            )
        else:
            attribute_lambda = attributes
        if isinstance(self, cls_type) and attribute_lambda(self):
            return self
        for child in self._children():
            try:
                return child.find(cls_type, attributes)
            except ValueError:
                pass
        msg = f"Object not found: {cls_type}"
        raise ValueError(msg)

    def model_dump_json(self, **kwargs: Any) -> str:
        return super().model_dump_json(round_trip=True, exclude_unset=True, **kwargs)

    def _children(self) -> list["LayoutNode"]:
        ret: list[LayoutNode] = []
        for attr in dir(self):
            if attr.startswith("_"):
                continue
            child_candidate: list[Any] | dict[str, Any] | LayoutNode | int | str = getattr(self, attr)
            if isinstance(child_candidate, list):
                ret.extend(val for val in child_candidate if isinstance(val, LayoutNode))
            elif isinstance(child_candidate, dict):
                ret.extend(val for val in child_candidate.values() if isinstance(val, LayoutNode))
            elif isinstance(child_candidate, LayoutNode):
                ret.append(child_candidate)
        return ret

    def __str__(self) -> str:
        if self._name_field is not None:
            name = getattr(self, self._name_field)
            if callable(name):
                name = name()
            return f"{self.__class__.__name__}({name})"
        return super().__str__()

    def get_lineage(
        self,
        lineage_type: Literal["children", "parents"],
        tabular_model: "BaseTabularModel",
    ) -> LineageNode:
        raise NotImplementedError

    def find_xpath(self, xpath: list[str | int]) -> "LayoutNode":
        return _find_xpath(self, xpath)

    def pprint(self, indent: int = 4) -> None:
        ret = self.model_dump_json(indent=indent)
        ret = ret.replace('"', "").replace(":", "=").replace("{", "(").replace("}", ")")
        ret = self.__class__.__name__ + ret
        print(ret)


def _find_xpath(val: LayoutNode | list[LayoutNode] | dict[str, LayoutNode] | str, xpath: list[str | int]) -> LayoutNode:
    if len(xpath) == 0:
        assert isinstance(val, LayoutNode)
        return val
    if isinstance(val, list):
        next_pos = xpath.pop(0)
        assert isinstance(next_pos, int)
        return _find_xpath(val[next_pos], xpath)
    if isinstance(val, dict):
        breakpoint()
    elif isinstance(val, LayoutNode):
        next_pos = xpath.pop(0)
        assert isinstance(next_pos, str)
        return _find_xpath(getattr(val, next_pos), xpath)
    msg = f"What? xpath={xpath}, val={val}"
    raise ValueError(msg)


def _set_parents(
    base: list[LayoutNode] | dict[str, LayoutNode] | LayoutNode | int | str,
    last_parent: "LayoutNode",
    curr_xpath: list[str | int],
) -> None:
    if isinstance(base, list):
        for i, val in enumerate(base):
            _set_parents(val, last_parent, [*curr_xpath, i])
    elif isinstance(base, dict):
        for key, val in base.items():
            _set_parents(val, last_parent, [*curr_xpath, key])
    elif isinstance(base, LayoutNode):
        base._parent = last_parent
        base._xpath = curr_xpath

        for field_name in base.model_fields:
            if field_name.startswith("_"):
                continue
            field_value = getattr(base, field_name)
            if isinstance(field_value, list | dict | LayoutNode):
                _set_parents(field_value, base, [*curr_xpath, field_name])



import datetime
from typing import TYPE_CHECKING, Literal
from uuid import UUID, uuid4

from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasRenameRecord
from pbi_core.ssas.model_tables.enums import DataState
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

from .enums import HideMembers

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.level import Level
    from pbi_core.ssas.model_tables.table import Table
    from pbi_core.ssas.model_tables.variation import Variation


class Hierarchy(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/4eff6661-1458-4c5a-9875-07218f1458e5)
    """

    description: str | None = None
    display_folder: str | None = None
    hide_members: HideMembers
    hierarchy_storage_id: int
    is_hidden: bool
    name: str
    state: DataState
    table_id: int
    """A foreign key to the Table object the hierarchy is stored under"""

    lineage_tag: UUID = uuid4()
    source_lineage_tag: UUID = uuid4()

    modified_time: datetime.datetime
    refreshed_time: datetime.datetime
    """The last time the sources for this hierarchy were refreshed"""
    structure_modified_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.hierarchy)

    def table(self) -> "Table":
        return self.tabular_model.tables.find({"id": self.table_id})

    def levels(self) -> set["Level"]:
        return self.tabular_model.levels.find_all({"hierarchy_id": self.id})

    def variations(self) -> set["Variation"]:
        return self.tabular_model.variations.find_all({"default_hierarchy_id": self.id})

    @classmethod
    def _db_command_obj_name(cls) -> str:
        return "Hierarchies"

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(
                self,
                lineage_type,
                [level.get_lineage(lineage_type) for level in self.levels()]
                + [variation.get_lineage(lineage_type) for variation in self.variations()],
            )

        return LineageNode(
            self,
            lineage_type,
            [
                self.table().get_lineage(lineage_type),
            ],
        )



import datetime
from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

from .enums import MemberType

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.role import Role


class RoleMembership(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/dbecc1f4-142b-4765-8374-a4d4dc51313b)
    """

    identity_provider: str
    member_id: str
    member_name: str
    member_type: MemberType
    role_id: int

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.role_membership)

    def role(self) -> "Role":
        return self.tabular_model.roles.find(self.role_id)



from enum import IntEnum


class ImpersonationMode(IntEnum):
    DEFAULT = 1
    IMPERSONATE_ACCOUNT = 2
    IMPERSONATE_ANONYMOUS = 3
    IMPERSONATE_CURRENT_USER = 4
    IMPERSONATE_SERVICE_ACCOUNT = 5
    IMPERSONATE_UNATTENDED_ACCOUNT = 6


class DataSourceType(IntEnum):
    PROVIDER = 1
    STRUCTURED = 2


class Isolation(IntEnum):
    READ_COMMITTED = 1
    SNAPSHOT = 2



import pathlib
import shutil
from typing import TYPE_CHECKING, Any, TypedDict, cast

import bs4
from structlog import get_logger

from pbi_core.ssas.model_tables.base.batch import AlterCommand, Batch
from pbi_core.ssas.server.utils import COMMAND_TEMPLATES

logger = get_logger()
if TYPE_CHECKING:
    from _typeshed import StrPath

    from pbi_core.ssas.model_tables import (
        KPI,
        AlternateOf,
        Annotation,
        AttributeHierarchy,
        CalcDependency,
        CalculationGroup,
        CalculationItem,
        Column,
        ColumnPermission,
        Culture,
        DataSource,
        DetailRowDefinition,
        Expression,
        ExtendedProperty,
        FormatStringDefinition,
        Group,
        GroupByColumn,
        Hierarchy,
        Level,
        LinguisticMetadata,
        Measure,
        Model,
        ObjectTranslation,
        Partition,
        Perspective,
        PerspectiveColumn,
        PerspectiveHierarchy,
        PerspectiveMeasure,
        PerspectiveSet,
        PerspectiveTable,
        QueryGroup,
        RefreshPolicy,
        RelatedColumnDetail,
        Relationship,
        Role,
        RoleMembership,
        Set,
        Table,
        TablePermission,
        Variation,
    )
    from pbi_core.ssas.model_tables.base.base_ssas_table import SsasTable
    from pbi_core.ssas.model_tables.base.ssas_tables import SsasAlter
    from pbi_core.ssas.server.server import BaseServer, LocalServer


class Update(TypedDict):
    added: list["SsasTable"]
    updated: list["SsasAlter"]
    deleted: list[int]


class BaseTabularModel:
    db_name: str
    server: "BaseServer"
    model: "Model"
    alternate_ofs: "Group[AlternateOf]"
    annotations: "Group[Annotation]"
    """Notes that can be attached to a variety other SSAS objects"""

    attribute_hierarchies: "Group[AttributeHierarchy]"
    calc_dependencies: "Group[CalcDependency]"
    """Returns all (multi-generational) ancestors of a calculation element"""
    calculation_groups: "Group[CalculationGroup]"
    calculation_items: "Group[CalculationItem]"
    columns: "Group[Column]"
    """Columns include source (MQuery) and calculate columns from tables"""

    column_permissions: "Group[ColumnPermission]"
    cultures: "Group[Culture]"
    data_sources: "Group[DataSource]"
    detail_row_definitions: "Group[DetailRowDefinition]"
    expressions: "Group[Expression]"
    extended_properties: "Group[ExtendedProperty]"
    format_string_definitions: "Group[FormatStringDefinition]"
    group_by_columns: "Group[GroupByColumn]"
    hierarchies: "Group[Hierarchy]"
    kpis: "Group[KPI]"
    levels: "Group[Level]"
    linguistic_metadata: "Group[LinguisticMetadata]"
    measures: "Group[Measure]"
    object_translations: "Group[ObjectTranslation]"
    partitions: "Group[Partition]"
    """Partitions are generally accessed to edit the Power Query of a Table"""

    perspectives: "Group[Perspective]"
    perspective_columns: "Group[PerspectiveColumn]"
    perspective_hierarchies: "Group[PerspectiveHierarchy]"
    perspective_measures: "Group[PerspectiveMeasure]"
    perspective_sets: "Group[PerspectiveSet]"
    perspective_tables: "Group[PerspectiveTable]"
    query_groups: "Group[QueryGroup]"
    refresh_policies: "Group[RefreshPolicy]"
    related_column_details: "Group[RelatedColumnDetail]"
    relationships: "Group[Relationship]"
    roles: "Group[Role]"
    role_memberships: "Group[RoleMembership]"
    sets: "Group[Set]"
    tables: "Group[Table]"
    """This class contains the logical elements of a PowerBI table"""

    table_permissions: "Group[TablePermission]"
    variations: "Group[Variation]"

    # need to include changes to the model object itself
    # For each table of entities, maps the entity ID to it's remote "modified hash" value
    _remote_state: dict[str, dict[int, int]]

    def __init__(self, db_name: str, server: "BaseServer") -> None:
        self.db_name = db_name
        self.server = server
        self._remote_state = {}

    def save_pbix(self, path: "StrPath") -> None:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"TabularModel(db_name={self.db_name}, server={self.server})"

    def to_local(self, pbix_path: pathlib.Path) -> "LocalTabularModel":
        server = cast("LocalServer", self.server)
        return LocalTabularModel(self.db_name, server, pbix_path)

    def sync_from(self) -> None:
        """Pulls data from the SSAS instance to the Python instance.

        Pairs with the functions `sync_to` that update records in the
        SSAS instance with data from Python
        """
        from pbi_core.ssas.model_tables import FIELD_TYPES  # noqa: PLC0415
        from pbi_core.ssas.model_tables._group import Group  # noqa: PLC0415

        xml_schema = self.server.query_xml(COMMAND_TEMPLATES["discover_schema.xml"].render(db_name=self.db_name))
        schema = discover_xml_to_dict(xml_schema)
        for field_name, type_instance in FIELD_TYPES.items():
            if field_name == "model":
                obj = type_instance.model_validate({
                    **schema[type_instance._db_type_name()][0],
                    "_tabular_model": self,
                })
                setattr(self, field_name, obj)
            else:
                objects = Group([
                    type_instance.model_validate({**row, "_tabular_model": self})
                    for row in schema[type_instance._db_type_name()]
                ])
                setattr(self, field_name, objects)

                self._remote_state.setdefault(field_name, {})
                for obj in objects:
                    self._remote_state[field_name][obj.id] = obj.modification_hash()

    def sync_to(self) -> None:
        # like {"tables": {"updated": [UpdatedObj1]}}
        updated_objects: dict[str, Update] = {}
        for field_name, remote_objects in self._remote_state.items():
            field_updates: Update = {
                "updated": [],
                "deleted": [],
                "added": [],
            }
            current_objects: Group[SsasTable] = getattr(self, field_name)
            field_updates["deleted"] = [obj.id for obj in current_objects if obj.id not in remote_objects]
            for obj in current_objects:
                if obj.id not in self._remote_state[field_name]:
                    field_updates["added"].append(obj)
                if obj.modification_hash() != self._remote_state[field_name][obj.id]:
                    field_updates["updated"].append(obj)  # pyright: ignore[reportArgumentType]
            if field_updates["added"] or field_updates["updated"] or field_updates["deleted"]:
                updated_objects[field_name] = field_updates

        commands = [
            AlterCommand(
                self.db_name,
                {field_name: updates["updated"] for field_name, updates in updated_objects.items()},
            ),
        ]
        command_str = Batch(commands).render_xml()
        self.server.query_xml(command_str, db_name=self.db_name)

    @staticmethod
    def TABULAR_FIELDS() -> tuple[str, ...]:  # noqa: N802
        """Returns a list of all the field names for the SSAS tables in the tabular model.

        No calc_dependencies, it's not a real table but a view
        No model, since it's not a "real" table
        """
        return (
            "alternate_ofs",
            "annotations",
            "attribute_hierarchies",
            "calculation_groups",
            "calculation_items",
            "column_permissions",
            "columns",
            "cultures",
            "data_sources",
            "detail_row_definitions",
            "expressions",
            "extended_properties",
            "format_string_definitions",
            "group_by_columns",
            "hierarchies",
            "kpis",
            "levels",
            "linguistic_metadata",
            "measures",
            "object_translations",
            "partitions",
            "perspectives",
            "perspective_columns",
            "perspective_hierarchies",
            "perspective_measures",
            "perspective_sets",
            "perspective_tables",
            "query_groups",
            "refresh_policies",
            "related_column_details",
            "relationships",
            "role_memberships",
            "roles",
            "sets",
            "table_permissions",
            "tables",
            "variations",
        )


class LocalTabularModel(BaseTabularModel):
    pbix_path: pathlib.Path
    server: "LocalServer"  # type: ignore[assignment]

    def __init__(self, db_name: str, server: "LocalServer", pbix_path: pathlib.Path) -> None:
        self.pbix_path = pbix_path
        super().__init__(db_name, server)

    def save_pbix(self, path: "StrPath") -> None:
        shutil.copy(self.pbix_path, path)
        self.server.save_pbix(path, self.db_name)  # pyright: ignore reportAttributeAccessIssue  # the server is always a local server in this case


def discover_xml_to_dict(xml: bs4.BeautifulSoup) -> dict[str, list[dict[Any, Any]]]:
    """Converts the results of the Discover XML to a dictionary to make downstream transformations more convenient."""
    assert xml.results is not None
    results = cast("list[bs4.element.Tag]", list(xml.results))
    results[-1]["name"] = "CalcDependency"
    ret = {
        cast("str", table["name"]): [
            {field.name: field.text for field in row if field.name is not None}  # pyright: ignore reportGeneralTypeIssues
            for row in table.find_all("row")
        ]
        for table in results
    }
    for i, row in enumerate(ret["CalcDependency"]):
        row["id"] = i
    return ret



from enum import IntEnum


class Alignment(IntEnum):
    DEFAULT = 1
    LEFT = 2
    RIGHT = 3
    CENTER = 4


class PropertyType(IntEnum):
    """Source: https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/00a9ec7a-5f4d-4517-8091-b370fe2dc18b."""

    INVALID = -1
    ALL = 1
    REGULAR = 2
    IMAGE = 3
    IMAGE_BMP = 4
    IMAGE_GIF = 5
    IMAGE_JPG = 6
    IMAGE_PNG = 7
    IMAGE_TIFF = 8
    IMAGE_URL = 9
    ID = 10
    RELATION_TO_PARENT = 11
    SEQUENCE = 12
    ORG_TITLE = 13
    CAPTION = 14
    SHORT_CAPTION = 15
    CAPTION_DESCRIPTION = 16
    CAPTION_ABBREVIATION = 17
    WEB_URL = 18
    WEB_HTML = 19
    WEB_XML_OR_XSL = 20
    WEBMAIL_ALIAS = 21
    ADDRESS = 22
    ADDRESS_STREET = 23
    ADDRESS_HOUSE = 24
    ADDRESS_CITY = 25
    ADDRESS_STATE_OR_PROVINCE = 26
    ADDRESS_ZIP = 27
    ADDRESS_QUARTER = 28
    ADDRESS_COUNTRY = 29
    ADDRESS_BUILDING = 30
    ADDRESS_ROOM = 31
    ADDRESS_FLOOR = 32
    ADDRESS_FAX = 33
    ADDRESS_PHONE = 34
    GEO_CENTROID_X = 35
    GEO_CENTROID_Y = 36
    GEO_CENTROID_Z = 37
    GEO_BOUNDARY_TOP = 38
    GEO_BOUNDARY_LEFT = 39
    GEO_BOUNDARY_BOTTOM = 40
    GEO_BOUNDARY_RIGHT = 41
    GEO_BOUNDARY_FRONT = 42
    GEO_BOUNDARY_REAR = 43
    GEO_BOUNDARY_POLYGON = 44
    PHYSICAL_SIZE = 45
    PHYSICAL_COLOR = 46
    PHYSICAL_WEIGHT = 47
    PHYSICAL_HEIGHT = 48
    PHYSICAL_WIDTH = 49
    PHYSICAL_DEPTH = 50
    PHYSICAL_VOLUME = 51
    PHYSICAL_DENSITY = 52
    PERSON_FULL_NAME = 53
    PERSON_FIRST_NAME = 54
    PERSON_LAST_NAME = 55
    PERSON_MIDDLE_NAME = 56
    PERSON_DEMOGRAPHIC = 57
    PERSON_CONTACT = 58
    QTY_RANGE_LOW = 59
    QTY_RANGE_HIGH = 60
    FORMATTING_COLOR = 61
    FORMATTING_ORDER = 62
    FORMATTING_FONT = 63
    FORMATTING_FONT_EFFECTS = 64
    FORMATTING_FONT_SIZE = 65
    FORMATTING_SUBTOTAL = 66
    DATE = 67
    DATE_START = 68
    DATE_ENDED = 69
    DATE_CANCELED = 70
    DATE_MODIFIED = 71
    DATE_DURATION = 72
    VERSION = 73
    YEARS = 74
    QUARTERS = 75
    MONTHS = 76
    WEEKS = 77
    DAYS = 78
    HOURS = 79
    MINUTES = 80
    SECONDS = 81
    UNDEFINED_TIME = 82
    ORGANIZATIONAL_UNIT = 83
    BOM_RESOURCE = 84
    QUANTITATIVE = 85
    ACCOUNT = 86
    CUSTOMERS = 87
    CUSTOMER_GROUP = 88
    CUSTOMER_HOUSEHOLD = 89
    PRODUCT = 90
    PRODUCT_GROUP = 91
    SCENARIO = 92
    UTILITY = 93
    PERSON = 94
    COMPANY = 95
    CURRENCY_SOURCE = 96
    CURRENCY_DESTINATION = 97
    CHANNEL = 98
    REPRESENTATIVE = 99
    PROMOTION = 100
    CONTINENT = 101
    REGION = 102
    COUNTRY = 103
    STATE_OR_PROVINCE = 104
    COUNTY = 105
    CITY = 106
    POSTAL_CODE = 107
    POINT = 108
    ACCOUNT_TYPE = 109
    ACCOUNT_NAME = 110
    ACCOUNT_NUMBER = 111
    PROJECT_NAME = 112
    PROJECT_CODE = 113
    PROJECT_START_DATE = 114
    PROJECT_END_DATE = 115
    PROJECT_COMPLETION = 116
    CURRENCY_NAME = 117
    CURRENCY_IS_O_CODE = 118
    PERCENT_OWNERSHIP = 119
    PERCENT_VOTERIGHT = 120
    PROJECT = 121
    RATE_TYPE = 122
    RATE = 123
    PRODUCT_SKU = 124
    PRODUCT_CATEGORY = 125
    PRODUCT_BRAND = 126
    DELETED_FLAG = 127
    SCD_STATUS = 128
    SCD_END_DATE = 129
    SCD_ORIGINAL_ID = 130
    SCD_START_DATE = 131
    DAY_OF_MONTH_OR_PERIOD = 132
    WEEK_OF_QUARTER = 133
    WEEK_OF_MONTH_OR_PERIOD = 134
    MONTH_OR_PERIOD_OF_QUARTER = 135
    MONTH_OR_PERIOD_OF_YEAR = 136
    TRIMESTERS = 137
    HALFYEARS = 138
    TENDAYS = 139
    DAY_OF_WEEK = 140
    DAY_OF_TENDAYS = 141
    DAY_OF_MONTH = 142
    DAY_OF_QUARTER = 143
    DAY_OF_TRIMESTER = 144
    DAY_OF_HALFYEAR = 145
    DAY_OF_YEAR = 146
    WEEK_OF_YEAR = 147
    TENDAY_OF_MONTH = 148
    TENDAY_OF_QUARTER = 149
    TENDAY_OF_TRIMESTER = 150
    TENDAY_OF_HALFYEAR = 151
    TENDAY_OF_YEAR = 152
    MONTH_OF_TRIMESTER = 153
    MONTH_OF_QUARTER = 154
    MONTH_OF_HALFYEAR = 155
    MONTH_OF_YEAR = 156
    TRIMESTER_OF_YEAR = 157
    QUARTER_OF_HALFYEAR = 158
    QUARTER_OF_YEAR = 159
    HALFYEAR_OF_YEAR = 160
    FISCAL_DATE = 161
    FISCAL_DAY_OF_WEEK = 162
    FISCAL_DAY_OF_MONTH = 163
    FISCAL_DAY_OF_QUARTER = 164
    FISCAL_DAY_OF_TRIMESTER = 165
    FISCAL_DAY_OF_HALFYEAR = 166
    FISCAL_DAY_OF_YEAR = 167
    FISCAL_WEEKS = 168
    FISCAL_WEEK_OF_YEAR = 169
    FISCAL_WEEK_OF_HALFYEAR = 170
    FISCAL_WEEK_OF_QUARTER = 171
    FISCAL_WEEK_OF_TRIMESTER = 172
    FISCAL_WEEK_OF_MONTH = 173
    FISCAL_MONTHS = 174
    FISCAL_MONTH_OF_TRIMESTER = 175
    FISCAL_MONTH_OF_QUARTER = 176
    FISCAL_MONTH_OF_HALFYEAR = 177
    FISCAL_MONTH_OF_YEAR = 178
    FISCAL_TRIMESTERS = 179
    FISCAL_TRIMESTER_OF_YEAR = 180
    FISCAL_QUARTERS = 181
    FISCAL_QUARTER_OF_YEAR = 182
    FISCAL_QUARTER_OF_HALFYEAR = 183
    FISCAL_HALFYEARS = 184
    FISCAL_HALFYEAR_OF_YEAR = 185
    FISCAL_YEARS = 186
    REPORTING_DATE = 187
    REPORTING_DAY_OF_WEEK = 188
    REPORTING_DAY_OF_MONTH = 189
    REPORTING_DAY_OF_QUARTER = 190
    REPORTING_DAY_OF_TRIMESTER = 191
    REPORTING_DAY_OF_HALFYEAR = 192
    REPORTING_DAY_OF_YEAR = 193
    REPORTING_WEEKS = 194
    REPORTING_WEEK_OF_YEAR = 195
    REPORTING_WEEK_OF_HALFYEAR = 196
    REPORTING_WEEK_OF_QUARTER = 197
    REPORTING_WEEK_OF_TRIMESTER = 198
    REPORTING_WEEK_OF_MONTH = 199
    REPORTING_MONTHS = 200
    REPORTING_MONTH_OF_TRIMESTER = 201
    REPORTING_MONTH_OF_QUARTER = 202
    REPORTING_MONTH_OF_HALFYEAR = 203
    REPORTING_MONTH_OF_YEAR = 204
    REPORTING_TRIMESTERS = 205
    REPORTING_TRIMESTER_OF_YEAR = 206
    REPORTING_QUARTERS = 207
    REPORTING_QUARTER_OF_YEAR = 208
    REPORTING_QUARTER_OF_HALFYEAR = 209
    REPORTING_HALFYEARS = 210
    REPORTING_HALFYEAR_OF_YEAR = 211
    REPORTING_YEARS = 212
    MANUFACTURING_DATE = 213
    MANUFACTURING_DAY_OF_WEEK = 214
    MANUFACTURING_DAY_OF_MONTH = 215
    MANUFACTURING_DAY_OF_QUARTER = 216
    MANUFACTURING_DAY_OF_HALFYEAR = 217
    MANUFACTURING_DAY_OF_YEAR = 218
    MANUFACTURING_WEEKS = 219
    MANUFACTURING_WEEK_OF_YEAR = 220
    MANUFACTURING_WEEK_OF_QUARTER = 222
    MANUFACTURING_WEEK_OF_MONTH = 223
    MANUFACTURING_MONTHS = 224
    MANUFACTURING_MONTH_OF_QUARTER = 225
    MANUFACTURING_MONTH_OF_HALFYEAR = 226
    MANUFACTURING_MONTH_OF_YEAR = 227
    MANUFACTURING_TRIMESTERS = 228
    MANUFACTURING_TRIMESTER_OF_YEAR = 229
    MANUFACTURING_QUARTERS = 230
    MANUFACTURING_QUARTER_OF_YEAR = 231
    MANUFACTURING_QUARTER_OF_HALFYEAR = 232
    MANUFACTURING_HALFYEARS = 233
    MANUFACTURING_HALFYEAR_OF_YEAR = 234
    MANUFACTURING_YEARS = 235
    WINTER_SUMMER_SEASON = 236
    IS_HOLIDAY = 237
    IS_WEEKDAY = 238
    IS_WORKING_DAY = 239
    IS_PEAK_DAY = 240
    ISO8601_DATE = 241
    ISO8601_DAY_OF_WEEK = 242
    ISO8601_DAY_OF_YEAR = 243
    ISO8601_WEEKS = 244
    ISO8601_WEEK_OF_YEAR = 245
    ISO8601_YEARS = 246
    ROW_NUMBER = 247
    EXTENDED_TYPE = 248


class SummarizedBy(IntEnum):
    DEFAULT = 1
    _NONE = 2
    SUM = 3
    MIN = 4
    MAX = 5
    COUNT = 6
    AVERAGE = 7
    COUNT_DISTINCT = 8


class ColumnType(IntEnum):
    DATA = 1
    """A standard column"""

    CALCULATED = 2
    """A DAX-calculated column"""

    ROW_NUMBER = 3
    """Hidden row number column on each table"""

    CALCULATED_TABLE_COLUMN = 4


class EncodingHint(IntEnum):
    DEFAULT = 0
    HASH = 1
    VALUE = 2



from typing import TYPE_CHECKING, Literal

from pbi_core.lineage import LineageNode

from .helpers import HelpersMixin

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.attribute_hierarchy import AttributeHierarchy
    from pbi_core.ssas.model_tables.base import SsasTable
    from pbi_core.ssas.model_tables.measure import Measure

    from .column import Column


class DependencyMixin(HelpersMixin):
    explicit_name: str | None
    expression: str | int | None

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            children_nodes: list[Column | Measure | AttributeHierarchy] = [
                self.attribute_hierarchy(),
                *self.child_measures(),
                *self.sorting_columns(),
                *self.child_columns(),
            ]
            children_lineage = [p.get_lineage(lineage_type) for p in children_nodes if p is not None]
            return LineageNode(self, lineage_type, children_lineage)
        parent_nodes: list[SsasTable | None] = [
            self.table(),
            self.sort_by_column(),
            *self.parent_columns(),
            *self.parent_measures(),
        ]
        parent_lineage = [p.get_lineage(lineage_type) for p in parent_nodes if p is not None]
        return LineageNode(self, lineage_type, parent_lineage)

    def child_measures(self, *, recursive: bool = False) -> set["Measure"]:
        """Returns measures dependent on this Column."""
        object_type = self._column_type()
        dependent_measures = self.tabular_model.calc_dependencies.find_all({
            "referenced_object_type": object_type,
            "referenced_table": self.table().name,
            "referenced_object": self.explicit_name,
            "object_type": "MEASURE",
        })
        child_keys: list[tuple[str | None, str]] = [(m.table, m.object) for m in dependent_measures]
        full_dependencies = [m for m in self.tabular_model.measures if (m.table().name, m.name) in child_keys]

        if recursive:
            recursive_dependencies: set[Measure] = set()
            for dep in full_dependencies:
                if f"[{self.explicit_name}]" in str(dep.expression):
                    recursive_dependencies.add(dep)
                    recursive_dependencies.update(dep.child_measures(recursive=True))
            return recursive_dependencies

        return {x for x in full_dependencies if f"[{self.explicit_name}]" in str(x.expression)}

    def parent_measures(self, *, recursive: bool = False) -> set["Measure"]:
        """Returns measures this column is dependent on.

        Note:
            Calculated columns can use Measures too :(.

        """
        object_type = self._column_type()
        dependent_measures = self.tabular_model.calc_dependencies.find_all({
            "object_type": object_type,
            "table": self.table().name,
            "object": self.explicit_name,
            "referenced_object_type": "MEASURE",
        })
        parent_keys = [(m.referenced_table, m.referenced_object) for m in dependent_measures]
        full_dependencies = [m for m in self.tabular_model.measures if (m.table().name, m.name) in parent_keys]

        if recursive:
            recursive_dependencies: set[Measure] = set()
            for dep in full_dependencies:
                if f"[{dep.name}]" in str(self.expression):
                    recursive_dependencies.add(dep)
                    recursive_dependencies.update(dep.parent_measures(recursive=True))
            return recursive_dependencies

        return {x for x in full_dependencies if f"[{x.name}]" in str(self.expression)}

    def child_columns(self, *, recursive: bool = False) -> set["Column"]:
        """Returns columns dependent on this Column.

        Note:
            Only occurs when the dependent column is calculated (expression is not None).

        """
        object_type = self._column_type()
        dependent_measures = self.tabular_model.calc_dependencies.find_all({
            "referenced_object_type": object_type,
            "referenced_table": self.table().name,
            "referenced_object": self.explicit_name,
        })
        assert all(m.table is not None for m in dependent_measures)
        child_keys: list[tuple[str, str]] = [  # pyright: ignore reportAssignmentType
            (m.table, m.object) for m in dependent_measures if m.object_type in {"CALC_COLUMN", "COLUMN"}
        ]
        full_dependencies = [m for m in self.tabular_model.columns if (m.table().name, m.explicit_name) in child_keys]

        if recursive:
            recursive_dependencies: set[Column] = set()
            for dep in full_dependencies:
                if f"[{self.explicit_name}]" in str(dep.expression):
                    recursive_dependencies.add(dep)
                    recursive_dependencies.update(dep.child_columns(recursive=True))
            return recursive_dependencies

        return {x for x in full_dependencies if f"[{self.explicit_name}]" in str(x.expression)}

    def parent_columns(self, *, recursive: bool = False) -> set["Column"]:
        """Returns Columns this Column is dependent on.

        Note:
            Parent columns are non-empty only when the column is calculated.
            Columns defined by a PowerQuery import or DirectQuery do not have parent columns.

        """
        object_type = self._column_type()
        if object_type == "COLUMN":
            return set()
        dependent_measures = self.tabular_model.calc_dependencies.find_all({
            "object_type": object_type,
            "table": self.table().name,
            "object": self.explicit_name,
        })
        parent_keys = {
            (m.referenced_table, m.referenced_object)
            for m in dependent_measures
            if m.referenced_object_type in {"CALC_COLUMN", "COLUMN"}
        }
        full_dependencies = [c for c in self.tabular_model.columns if (c.table().name, c.explicit_name) in parent_keys]

        if recursive:
            recursive_dependencies: set[Column] = set()
            for dep in full_dependencies:
                if f"[{dep.explicit_name}]" in str(self.expression):
                    recursive_dependencies.add(dep)
                    recursive_dependencies.update(dep.parent_columns(recursive=True))
            return recursive_dependencies

        return {x for x in full_dependencies if f"[{x.explicit_name}]" in str(self.expression)}

    def parents(self, *, recursive: bool = False) -> "set[Column | Measure]":
        """Returns all columns and measures this Column is dependent on."""
        full_dependencies = self.parent_columns() | self.parent_measures()
        if recursive:
            recursive_dependencies: set[Column | Measure] = set()
            for dep in full_dependencies:
                recursive_dependencies.add(dep)
                recursive_dependencies.update(dep.parents(recursive=True))
            return recursive_dependencies

        return full_dependencies

    def children(self, *, recursive: bool = False) -> "set[Column | Measure]":
        """Returns all columns and measures dependent on this Column."""
        full_dependencies = self.child_columns() | self.child_measures()
        if recursive:
            recursive_dependencies: set[Column | Measure] = set()
            for dep in full_dependencies:
                recursive_dependencies.add(dep)
                recursive_dependencies.update(dep.children(recursive=True))
            return recursive_dependencies
        return full_dependencies



from .server import BaseServer, LocalServer, get_or_create_local_server, list_local_servers, terminate_all_local_servers
from .tabular_model import BaseTabularModel, LocalTabularModel

__all__ = [
    "BaseServer",
    "BaseTabularModel",
    "LocalServer",
    "LocalTabularModel",
    "get_or_create_local_server",
    "list_local_servers",
    "terminate_all_local_servers",
]



from enum import Enum, IntEnum
from typing import TYPE_CHECKING, Annotated, Any, Literal

from pydantic import Discriminator, Json, Tag

from pbi_core.lineage.main import LineageNode
from pbi_core.static_files.model_references import ModelColumnReference, ModelMeasureReference

from ._base_node import LayoutNode
from .condition import Condition
from .filters import From as FromType
from .filters import PrototypeQuery, PrototypeQueryResult, VisualFilter
from .selector import Selector
from .sources import Source
from .visuals.base import FilterSortOrder, ProjectionConfig, PropertyDef
from .visuals.main import Visual
from .visuals.properties.base import Expression

if TYPE_CHECKING:
    from pbi_core.ssas.server import LocalTabularModel
    from pbi_core.ssas.server.tabular_model.tabular_model import BaseTabularModel

    from .section import Section


from .expansion_state import ExpansionState
from .performance import Performance, get_performance


class BackgroundProperties(LayoutNode):
    show: Expression | None = None
    transparency: Expression | None = None


class Background(LayoutNode):
    properties: BackgroundProperties


class SingleVisualGroupProperties(LayoutNode):
    background: list[Background] | None = None


class SingleVisualGroup(LayoutNode):
    displayName: str
    groupMode: int
    objects: SingleVisualGroupProperties | None = None
    isHidden: bool = False


class VisualHowCreated(Enum):
    INSERT_VISUAL_BUTTON = "InsertVisualButton"


class VisualLayoutInfoPosition(LayoutNode):
    x: float
    y: float
    z: float = 0.0  # z is not always present, default to 0
    width: float
    height: float
    tabOrder: int | None = None

    def __str__(self) -> str:
        return f"({self.x}, {self.y}, {self.z}, {self.width}, {self.height})"


class VisualLayoutInfo(LayoutNode):
    id: int
    position: VisualLayoutInfoPosition


class VisualConfig(LayoutNode):
    _parent: "VisualContainer"  # pyright: ignore reportIncompatibleVariableOverride=false
    _name_field = "name"

    layouts: list[VisualLayoutInfo] | None = None
    name: str | None = None
    parentGroupName: str | None = None
    singleVisualGroup: SingleVisualGroup | None = None
    singleVisual: Visual | None = None  # split classes to handle the other cases
    howCreated: VisualHowCreated | None = None


class ExecutionMetricsKindEnum(IntEnum):
    NA = 1


class EntityType(IntEnum):
    TABLE = 0


class FromEntity(LayoutNode):
    Name: str
    Entity: str
    Type: EntityType = EntityType.TABLE


class PrimaryProjections(LayoutNode):
    Projections: list[int]
    SuppressedProjections: list[int] | None = None
    Subtotal: int | None = None
    Aggregates: list["QueryBindingAggregates"] | None = None
    ShowItemsWithNoData: list[int] | None = None


class Level(LayoutNode):
    Expressions: list[Source]
    Default: int


class InstanceChild(LayoutNode):
    Values: list[Source]
    Children: list["InstanceChild"] | None = None
    WindowExpansionInstanceWindowValue: list[int] | None = None  # never seen the element


class Instance(LayoutNode):
    Children: list[InstanceChild]
    WindowExpansionInstanceWindowValue: list[int] | None = None  # never seen the element
    Values: list[Source] | None = None


class BindingExpansion(LayoutNode):
    From: list[FromEntity]
    Levels: list[Level]
    Instances: Instance


class Synch(LayoutNode):
    Groupings: list[int]


class BindingPrimary(LayoutNode):
    Groupings: list[PrimaryProjections]
    Expansion: BindingExpansion | None = None
    Synchronization: list[Synch] | None = None


class DataVolume(IntEnum):
    NA1 = 1
    NA2 = 2
    NA3 = 3
    NA = 4
    NA5 = 5
    NA6 = 6


class SampleDataReduction(LayoutNode):
    Sample: dict[str, int]


class WindowDataReduction(LayoutNode):
    Window: dict[str, int]


class TopDataReduction(LayoutNode):
    Top: dict[str, int]


class BottomDataReduction(LayoutNode):
    Bottom: dict[str, int]


class OverlappingPointsSample(LayoutNode):
    X: dict[str, int]
    Y: dict[str, int]


class OverlappingPointReduction(LayoutNode):
    OverlappingPointsSample: OverlappingPointsSample


class WindowExpansionType(LayoutNode):
    From: list[FromEntity]
    Levels: list[Level]
    WindowInstances: Instance

    def __str__(self) -> str:
        return f"WindowExpansionType(From={self.From}, Levels={self.Levels}, WindowInstances={self.WindowInstances})"


class TopNPerLevelDataReductionHelper(LayoutNode):
    Count: int
    WindowExpansion: WindowExpansionType


class TopNPerLevelDataReduction(LayoutNode):
    TopNPerLevel: TopNPerLevelDataReductionHelper


class BinnedLineSampleHelper(LayoutNode):
    PrimaryScalarKey: int | None = None
    Count: int | None = None
    WarningCount: int | None = None


class BinnedLineSample(LayoutNode):
    BinnedLineSample: BinnedLineSampleHelper


def get_reduction(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        mapper = {
            "Sample": "SampleDataReduction",
            "Window": "WindowDataReduction",
            "Top": "TopDataReduction",
            "Bottom": "BottomDataReduction",
            "OverlappingPointsSample": "OverlappingPointReduction",
            "TopNPerLevel": "TopNPerLevelDataReduction",
            "BinnedLineSample": "BinnedLineSample",
        }

        for key in v:
            if key in mapper:
                return mapper[key]
        msg = f"Unknown Filter: {v.keys()}"
        raise ValueError(msg)

    return v.__class__.__name__


PrimaryDataReduction = Annotated[
    Annotated[SampleDataReduction, Tag("SampleDataReduction")]
    | Annotated[WindowDataReduction, Tag("WindowDataReduction")]
    | Annotated[TopDataReduction, Tag("TopDataReduction")]
    | Annotated[BottomDataReduction, Tag("BottomDataReduction")]
    | Annotated[OverlappingPointReduction, Tag("OverlappingPointReduction")]
    | Annotated[TopNPerLevelDataReduction, Tag("TopNPerLevelDataReduction")]
    | Annotated[BinnedLineSample, Tag("BinnedLineSample")],
    Discriminator(get_reduction),
]


class Scope(LayoutNode):
    Algorithm: PrimaryDataReduction
    Scope: dict[str, list[int]]


class DataReductionType(LayoutNode):
    DataVolume: DataVolume
    Primary: PrimaryDataReduction | None = None
    Secondary: PrimaryDataReduction | None = None
    Intersection: PrimaryDataReduction | None = None
    Scoped: list[Scope] | None = None


class AggregateSourceScope(LayoutNode):
    PrimaryDepth: int


class AggregateSources2(LayoutNode):  # stupid name, but needs to be different from AggregateSources
    # This is a workaround for the fact that AggregateSources is already used in the QueryBindingAggregates class
    Min: dict[str, int] | None = None
    Max: dict[str, int] | None = None
    Scope: AggregateSourceScope | None = None
    RespectInstanceFilters: bool = False


class AggregateSources(LayoutNode):
    min: dict[str, int] | None = None
    max: dict[str, int] | None = None


class QueryBindingAggregates(LayoutNode):
    Aggregations: list[AggregateSources2]
    Select: int


class Highlight(LayoutNode):
    # TODO: merge with VisualFilterExpression. For some reason,
    # pydantic thinks From should be None when using the visal filter expression
    Version: int | None = None
    From: list[FromType] | None = None
    Where: list[Condition]


class QueryBinding(LayoutNode):
    IncludeEmptyGroups: bool = False
    Primary: BindingPrimary
    Secondary: BindingPrimary | None = None
    Projections: list[int] = []
    DataReduction: DataReductionType | None = None
    Aggregates: list[QueryBindingAggregates] | None = None
    SuppressedJoinPredicates: list[int] | None = None
    Highlights: list[Highlight] | None = None
    Version: int


class QueryCommand1(LayoutNode):
    ExecutionMetricsKind: ExecutionMetricsKindEnum = ExecutionMetricsKindEnum.NA
    Query: PrototypeQuery
    Binding: QueryBinding | None = None


class QueryCommand2(LayoutNode):
    SemanticQueryDataShapeCommand: QueryCommand1


def get_query_command(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "SemanticQueryDataShapeCommand" in v:
            return "QueryCommand2"
        if "ExecutionMetricsKind" in v:
            return "QueryCommand1"
        msg = f"Unknown Filter: {v.keys()}"
        raise ValueError(msg)
    return v.__class__.__name__


QueryCommand = Annotated[
    Annotated[QueryCommand1, Tag("QueryCommand1")] | Annotated[QueryCommand2, Tag("QueryCommand2")],
    Discriminator(get_query_command),
]


class Query(LayoutNode):
    Commands: list[QueryCommand]

    def get_ssas_elements(self) -> set[ModelColumnReference | ModelMeasureReference]:
        """Returns the SSAS elements (columns and measures) this query is directly dependent on."""
        ret: set[ModelColumnReference | ModelMeasureReference] = set()
        for command in self.Commands:
            if isinstance(command, QueryCommand1):
                ret.update(command.Query.get_ssas_elements())
            elif isinstance(command, QueryCommand2):
                ret.update(command.SemanticQueryDataShapeCommand.Query.get_ssas_elements())
        return ret

    def get_prototype_queries(self) -> list[PrototypeQuery]:
        ret: list[PrototypeQuery] = []
        for command in self.Commands:
            if isinstance(command, QueryCommand1):
                ret.append(command.Query)
            elif isinstance(command, QueryCommand2):
                ret.append(command.SemanticQueryDataShapeCommand.Query)
        return ret


class Split(LayoutNode):
    # TODO: these strings are all stringy ints
    selects: dict[str, bool]


class KPI(LayoutNode):
    graphic: str
    normalizedFiveStateKpiRange: bool


class Restatement(LayoutNode):
    Restatement: str
    Name: str
    Type: int  # TODO: make enum
    DataCategory: int | None = None  # TODO: make enum
    Format: str | None = None
    kpi: KPI | None = None


class QueryMetadataFilter(LayoutNode):
    type: int | None = None  # TODO: make enum
    expression: Source | None = None


class QueryMetadata(LayoutNode):
    Select: list[Restatement]
    Filters: list[QueryMetadataFilter] | None = None


class DataRole(LayoutNode):
    Name: str
    Projection: int
    isActive: bool


class DataTransformVisualElement(LayoutNode):
    DataRoles: list[DataRole]


class DataTransformSelectType(LayoutNode):
    category: str | None = None
    underlyingType: int | None = None  # TODO: make enum


class ColumnFormattingDataBars(LayoutNode):
    metadata: str


class ColumnFormatting(LayoutNode):
    dataBars: list[ColumnFormattingDataBars]


class Title(LayoutNode):
    text: list[None]


class Values(LayoutNode):
    fontColor: list[Selector]


class RelatedObjects(LayoutNode):
    columnFormatting: ColumnFormatting | None = None
    title: Title | None = None
    values: Values | None = None


class DataTransformSelect(LayoutNode):
    displayName: str | None = None
    format: str | None = None
    queryName: str
    roles: dict[str, bool] | None = None
    sort: int | None = None  # TODO: make enum
    aggregateSources: AggregateSources | None = None
    sortOrder: FilterSortOrder = FilterSortOrder.NA
    type: DataTransformSelectType | None = None
    expr: Source
    relatedObjects: RelatedObjects | None = None
    kpi: KPI | None = None


class DataTransform(LayoutNode):
    objects: dict[str, list[PropertyDef]] | None = None
    projectionOrdering: dict[str, list[int]]
    projectionActiveItems: dict[str, list[ProjectionConfig]] | None = None
    splits: list[Split] | None = None
    queryMetadata: QueryMetadata | None = None
    visualElements: list[DataTransformVisualElement] | None = None
    selects: list[DataTransformSelect]
    expansionStates: list[ExpansionState] | None = None


class VisualContainer(LayoutNode):
    """A Container for visuals in a report page.

    Generally, this is 1-1 with a real visual (bar chart, etc.), but can contain 0 (text boxes) or >1.
    It's at this level that the report connects with the SSAS model to get data for each visual.
    """

    _parent: "Section"  # pyright: ignore reportIncompatibleVariableOverride=false
    _name_field = "name"

    x: float
    y: float
    z: float
    width: float
    height: float
    tabOrder: int | None = None
    dataTransforms: Json[DataTransform] | None = None
    query: Json[Query] | None = None
    queryHash: int | None = None
    filters: Json[list[VisualFilter]] = []
    config: Json[VisualConfig]

    id: int | None = None

    def pbi_core_id(self) -> str:
        """Returns a unique identifier for the visual container.

        Seems to stay the same after edits and after copies of the visual are made (the copies are
            assigned new, unrelated IDs). In some cases, it appears that the name is only unique within a section.

        Raises:
            ValueError: If the visual container does not have an ID or a name defined in the config.

        """
        if self.id is not None:
            return str(self.id)
        if self.config.name is not None:
            return self.config.name
        msg = "VisualContainer must have an id or a name in config"
        raise ValueError(msg)

    def pbi_core_name(self) -> str:
        viz = self.config.singleVisual
        assert viz is not None
        return viz.visualType

    def name(self) -> str | None:
        if self.config.singleVisual is not None:
            return f"{self.config.singleVisual.visualType}(x={round(self.x, 2)}, y={round(self.y, 2)}, z={round(self.z, 2)})"  # noqa: E501
        return None

    def _get_data_command(self) -> PrototypeQuery | None:
        if self.query is None:
            return None
        if len(self.query.Commands) == 0:
            return None

        if len(self.query.Commands) > 1:
            msg = "Cannot get data for multiple commands"
            raise NotImplementedError(msg)

        query_command = self.query.Commands[0]
        if isinstance(query_command, QueryCommand1):
            query = query_command.Query
        else:
            query = query_command.SemanticQueryDataShapeCommand.Query
        return query

    def get_data(self, model: "LocalTabularModel") -> PrototypeQueryResult | None:
        """Gets data that would populate this visual from the SSAS DB.

        Uses the PrototypeQuery found within query to generate a DAX statement that then gets passed to SSAS.

        Returns None for non-data visuals such as static text boxes

        """
        query = self._get_data_command()
        if query is None:
            return None
        return query.get_data(model)

    def get_performance(self, model: "LocalTabularModel") -> Performance:
        """Calculates various metrics on the speed of the visual.

        Current Metrics:
            Total Seconds to Query
            Total Rows Retrieved

        Raises:
            TypeError: If the visual does not have a query command.

        """
        command = self._get_data_command()
        if command is None:
            msg = "Cannot get performance for a visual without a query command"
            raise TypeError(msg)
        return get_performance(model, [command.get_dax(model).dax])[0]

    def get_ssas_elements(self) -> set[ModelColumnReference | ModelMeasureReference]:
        """Returns the SSAS elements (columns and measures) this visual is directly dependent on."""
        ret: set[ModelColumnReference | ModelMeasureReference] = set()
        if self.config.singleVisual is not None:
            ret.update(self.config.singleVisual.get_ssas_elements())
        if self.query is not None:
            ret.update(self.query.get_ssas_elements())
        for f in self.filters:
            ret.update(f.get_ssas_elements())
        return ret

    # TODO: replace ._parent with a get_parent method
    def get_lineage(
        self,
        lineage_type: Literal["children", "parents"],
        tabular_model: "BaseTabularModel",
    ) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)

        viz_entities = self.get_ssas_elements()
        page_filters = self._parent.get_ssas_elements(include_visuals=False)
        report_filters = self._parent._parent.get_ssas_elements(include_sections=False)
        entities = viz_entities | page_filters | report_filters
        children_nodes = [ref.to_model(tabular_model) for ref in entities]

        children_lineage = [p.get_lineage(lineage_type) for p in children_nodes if p is not None]
        return LineageNode(self, lineage_type, children_lineage)



from typing import Annotated, Any

from pydantic import Discriminator, Tag

from pbi_core.static_files.layout._base_node import LayoutNode

from .base import SourceExpression, SourceRef


class PropertyVariationSource(LayoutNode):
    Expression: SourceRef
    Name: str
    Property: str

    def column(self) -> str:
        return self.Property


class _PropertyVariationSourceHelper(LayoutNode):
    PropertyVariationSource: PropertyVariationSource

    def table(self) -> str:
        return self.PropertyVariationSource.Expression.table()

    def column(self) -> str:
        return self.PropertyVariationSource.column()


def get_hierarchy_expression_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "PropertyVariationSource" in v:
            return "_PropertyVariationSourceHelper"
        if "SourceRef" in v:
            return "SourceRef"
        if "Property" in v:
            return "SourceExpression"
        raise ValueError
    return v.__class__.__name__


ConditionType = Annotated[
    Annotated[SourceExpression, Tag("SourceExpression")]
    | Annotated[_PropertyVariationSourceHelper, Tag("_PropertyVariationSourceHelper")]
    | Annotated[SourceRef, Tag("SourceRef")],
    Discriminator(get_hierarchy_expression_type),
]


class _HierarchySourceHelper(LayoutNode):
    Expression: ConditionType
    Hierarchy: str | None = None


class HierarchySource(LayoutNode):
    Hierarchy: _HierarchySourceHelper


class _HierarchyLevelSourceHelper(LayoutNode):
    _parent: "HierarchyLevelSource"  # pyright: ignore reportIncompatibleVariableOverride=false

    Expression: HierarchySource
    Level: str | None = None


class HierarchyLevelSource(LayoutNode):
    HierarchyLevel: _HierarchyLevelSourceHelper
    Name: str | None = None
    NativeReferenceName: str | None = None

    def __repr__(self) -> str:
        table = self.HierarchyLevel.Expression.Hierarchy.Expression.table()
        if isinstance(self.HierarchyLevel.Expression.Hierarchy.Expression, SourceRef):
            column = self.HierarchyLevel.Expression.Hierarchy.Hierarchy
        else:
            column = self.HierarchyLevel.Expression.Hierarchy.Expression.column()
        level = self.HierarchyLevel.Level
        return f"HierarchyLevelSource({table}.{column}.{level})"



import datetime
from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasRenameRecord
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.model import Model


class Perspective(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/8bbe314e-f308-4732-875c-9530a1b0fe95)
    """

    description: int
    model_id: int
    name: str

    modified_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.perspective)

    def model(self) -> "Model":
        return self.tabular_model.model



from typing import Any, Literal, Protocol

from pbi_mermaid import Link, MermaidDiagram, Node
from pbi_mermaid.node import NodeShape

"""
#ffffcc, #ffcc99, #ffcccc,
#ff99cc, #ffccff, #cc99ff,
#ccccff, #99ccff, #ccffff,
#99ffcc, #ccffcc, #ccff99,
"""

CLASS_STYLES = {
    # SSAS Stuff
    "AttributeHierarchy": "fill:#ffffcc,stroke:#333,stroke-width:1px",
    "Column": "fill:#ffcc99,stroke:#333,stroke-width:1px",
    "Culture": "fill:#ffcccc,stroke:#333,stroke-width:1px",
    "LinguisticMetadata": "fill:#ff99cc,stroke:#333,stroke-width:1px",
    "Measure": "fill:#ffccff,stroke:#333,stroke-width:1px",
    "Table": "fill:#cc99ff,stroke:#333,stroke-width:1px",
    "Partition": "fill:#ccccff,stroke:#333,stroke-width:1px",
    "Model": "fill:#99ccff,stroke:#333,stroke-width:1px",
    # Visuals
    "BarChart": "fill:#ccffff,stroke:#333,stroke-width:1px",
    "VisualContainer": "fill:#99ffcc,stroke:#333,stroke-width:1px",
    "Section": "fill:#ccffcc,stroke:#333,stroke-width:1px",
    "Layout": "fill:#ccff99,stroke:#333,stroke-width:1px",
}


class LineageProtocol(Protocol):
    id: str

    def pbi_core_name(self) -> str:
        """Returns the name displayed in the PBIX report."""
        raise NotImplementedError


class LineageNode:
    """Class used to track DAX dependencies from visuals to measures to PowerQueries."""

    value: Any
    relatives: list["LineageNode"]
    lineage_type: Literal["children", "parents"]

    def __init__(
        self,
        value: Any,
        lineage_type: Literal["children", "parents"],
        relatives: list["LineageNode"] | None = None,
    ) -> None:
        """Initialize."""
        self.value = value
        self.relatives = relatives or []
        self.lineage_type = lineage_type

    @staticmethod
    def _create_node(value: LineageProtocol) -> Node:
        return Node(
            id=f"{value.__class__.__name__}-{value.id}",
            content=value.pbi_core_name(),
            style=CLASS_STYLES[value.__class__.__name__],
            shape=NodeShape.round_edge,
        )

    def _to_mermaid_helper(self, node: Node) -> tuple[list[Node], list[Link]]:
        nodes: list[Node] = [node]
        links: list[Link] = []
        for relative in self.relatives:
            child_node = self._create_node(relative.value)
            child_nodes, child_links = relative._to_mermaid_helper(child_node)
            links.append(Link(nodes[0], child_node))
            links.extend(child_links)
            nodes.extend(child_nodes)
        return nodes, links

    def to_mermaid(self) -> MermaidDiagram:
        base_node = self._create_node(self.value)
        nodes, links = self._to_mermaid_helper(base_node)
        return MermaidDiagram(title="Lineage Chart", nodes=nodes, links=links)



from enum import IntEnum


class PolicyType(IntEnum):
    BASIC = 0


class Granularity(IntEnum):
    INVALID = -1
    DAY = 0
    MONTH = 1
    QUARTER = 2
    YEAR = 3


class RefreshMode(IntEnum):
    IMPORT = 0
    HYBRID = 1



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode

from .base import BaseVisual
from .properties.base import Expression


class FillPropertiesHelper(LayoutNode):
    fillColor: Expression | None = None
    show: Expression | None = None
    transparency: Expression | None = None


class FillProperties(LayoutNode):
    properties: FillPropertiesHelper


class GeneralPropertiesHelper(LayoutNode):
    shapeType: Expression | None = None


class GeneralProperties(LayoutNode):
    properties: GeneralPropertiesHelper


class LinePropertiesHelper(LayoutNode):
    lineColor: Expression | None = None
    roundEdge: Expression | None = None
    transparency: Expression | None = None
    weight: Expression | None = None


class LineProperties(LayoutNode):
    properties: LinePropertiesHelper


class RotationPropertiesHelper(LayoutNode):
    angle: Expression | None = None


class RotationProperties(LayoutNode):
    properties: RotationPropertiesHelper


class BasicShapeProperties(LayoutNode):
    fill: list[FillProperties] | None = None
    general: list[GeneralProperties] | None = None
    line: list[LineProperties] | None = None
    rotation: list[RotationProperties] | None = None


class BasicShape(BaseVisual):
    visualType: str = "basicShape"
    model_config = ConfigDict(extra="forbid")

    drillFilterOtherVisuals: bool = True
    objects: BasicShapeProperties | None = None



from enum import IntEnum
from typing import Annotated, Any

from pydantic import Discriminator, Tag

from pbi_core.static_files.layout._base_node import LayoutNode

from .aggregation import AggregationSource, DataSource


def get_expression_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "Aggregation" in v:
            return "AggregationSource"
        if any(c in v for c in ("Column", "Measure", "HierarchyLevel")):
            return "DataSource"
        raise TypeError(v)
    return v.__class__.__name__


Expression = Annotated[
    Annotated[DataSource, Tag("DataSource")] | Annotated[AggregationSource, Tag("AggregationSource")],
    Discriminator(get_expression_type),
]


class AllRolesRef(LayoutNode):
    AllRolesRef: bool = True  # no values have been seen in this field


class ScopedEval2(LayoutNode):
    Expression: Expression
    Scope: list[AllRolesRef]


class ScopedEval(LayoutNode):
    ScopedEval: ScopedEval2


class ArithmeticOperator(IntEnum):
    DIVIDE = 3


class _ArithmeticSourceHelper(LayoutNode):
    Left: Expression
    Right: ScopedEval
    Operator: ArithmeticOperator


class ArithmeticSource(LayoutNode):
    Arithmetic: _ArithmeticSourceHelper
    Name: str | None = None

    def get_sources(self) -> list[DataSource]:
        left = self.Arithmetic.Left
        if isinstance(left, AggregationSource):
            return left.get_sources()
        return [left]



from enum import IntEnum


class CommandBeginColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    STARTTIME = 3
    SESSITON = 8
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    NTDOMAINNAME = 33
    CLIENTPROCESSID = 36
    APPLICATIONNAME = 37
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERNAME = 43
    REQUESTPARAMETERS = 44
    REQUESTPROPERTIES = 45
    ACTIVITYID = 46
    REQUESTID = 47
    USEROBJECTID = 51
    APPLICATIONCONTEXT = 52
    DATABASEFRIENDLYNAME = 54
    IDENTITY = 55


class CommandEndColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    SESSITON = 8
    INTEGERDATA = 10
    SEVERITY = 22
    SUCCESS = 23
    ERROR = 24
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    NTDOMAINNAME = 33
    CLIENTPROCESSID = 36
    APPLICATIONNAME = 37
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERNAME = 43
    ACTIVITYID = 46
    REQUESTID = 47
    ERRORTYPE = 49
    USEROBJECTID = 51
    APPLICATIONCONTEXT = 52
    DATABASEFRIENDLYNAME = 54
    IDENTITY = 55




from pydantic_extra_types.color import Color

from ._base import BaseFileModel

base_val = bool | int | str


class TextClass(BaseFileModel):
    fontSize: int
    fontFace: str
    color: Color


class TextClasses(BaseFileModel):
    callout: TextClass
    title: TextClass
    header: TextClass
    label: TextClass


class ColorName(BaseFileModel):
    color: Color


class ColorThemeData(BaseFileModel):
    solid: ColorName


class Theme(BaseFileModel):
    """A class mapping the fields of the Theme JSON.

    Documented `here <https://learn.microsoft.com/en-us/power-bi/create-reports/desktop-report-themes#set-theme-colors>`
    """

    name: str
    dataColors: list[str] = []
    foreground: str | None = None
    foregroundNeutralSecondary: Color | None = None
    foregroundNeutralTertiary: Color | None = None
    background: Color | None = None
    backgroundLight: Color | None = None
    backgroundNeutral: Color | None = None
    tableAccent: Color | None = None
    good: Color | None = None
    neutral: Color | None = None
    bad: Color | None = None
    maximum: Color | None = None
    center: Color | None = None
    minimum: Color | None = None
    null: Color | None = None
    hyperlink: Color | None = None
    visitedHyperlink: Color | None = None

    textClasses: TextClasses | None = None
    visualStyles: dict[str, dict[str, dict[str, list[dict[str, bool | int | str | ColorThemeData]]]]]



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class GeneralPropertiesHelper(LayoutNode):
    imageUrl: Expression | None = None


class GeneralProperties(LayoutNode):
    properties: GeneralPropertiesHelper


class ImageScalingPropertiesHelper(LayoutNode):
    imageScalingType: Expression | None = None


class ImageScalingProperties(LayoutNode):
    properties: ImageScalingPropertiesHelper


class ImageProperties(LayoutNode):
    general: list[GeneralProperties] | None = None
    imageScaling: list[ImageScalingProperties] | None = None


class Image(BaseVisual):
    visualType: str = "image"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: ImageProperties | None = None



from enum import Enum
from typing import TYPE_CHECKING, Annotated, Any

from pydantic import Discriminator, Tag

from ._base_node import LayoutNode
from .expansion_state import ExpansionStateLevel, ExpansionStateRoot
from .filters import BookmarkFilter, CachedDisplayNames, Direction, FilterExpressionMetadata, HighlightScope, Orderby
from .selector import Selector
from .sources import Source
from .visuals.base import PropertyDef
from .visuals.properties.base import Expression

if TYPE_CHECKING:
    from .layout import Layout
    from .section import Section
    from .visuals.base import BaseVisual


class BookmarkFilters(LayoutNode):
    byExpr: list[BookmarkFilter] = []
    """Filter containers that will be identified by expression."""
    byType: list[BookmarkFilter] = []
    """Filter containers that will be identified by type."""
    byName: dict[str, BookmarkFilter] = {}
    """Filter containers that will be identified by name."""
    byTransientState: list[BookmarkFilter] = []
    """Filter containers that are transient"""


class HighlightSelection(LayoutNode):
    dataMap: dict[str, list[HighlightScope]]
    metadata: list[str] | None = None


class Highlight(LayoutNode):
    selection: list[HighlightSelection]
    filterExpressionMetadata: FilterExpressionMetadata | None = None


class DisplayMode(Enum):
    maximize = "maximize"
    """Visual is shown full screen."""
    spotlight = "spotlight"
    """Visual is spotlighted and other visuals on the page are dimmed."""
    elevation = "elevation"
    """Visual is shown with an elevation."""
    hidden = "hidden"
    """Visual is hidden."""


class DataTableOptions(Enum):
    accessible = "accessible"
    normal = "normal"


class MaximizedOptions(LayoutNode):
    dataTable: DataTableOptions


class Display(LayoutNode):
    mode: DisplayMode
    maximizedOptions: MaximizedOptions | None = None


class Remove(LayoutNode):
    object: str
    property: str
    selector: Selector | None = None


class BookmarkPartialVisualObject(LayoutNode):
    merge: dict[str, list[PropertyDef]] | None = None
    remove: list[Remove] | None = None


class ExpansionState(LayoutNode):
    roles: list[str]
    levels: list[ExpansionStateLevel]
    root: ExpansionStateRoot


class Parameter(LayoutNode):
    expr: Source
    """Parameter expression."""
    index: int
    """Index from which parameter starts in the visual projection."""
    length: int
    """Number of fields created by the parameter."""
    sortDirection: Direction | None = None
    """If the sort direction is set, the visual is sorted by this field parameter.
    The implication of a visual being sorted by a field parameter is as follows during parameter resolution:
    - If none of the newly projected fields exist in the sort list, apply the parameter sort direction to the first
        projected field and add it to the end of the sort list.
    - If all the projected fields in the sort list have the opposite sort direction as the parameter's sort direction,
        flip the parameter's sort direction."""


class BookmarkPartialVisual(LayoutNode):
    visualType: str
    """Name of visual"""
    autoSelectVisualType: bool = False
    """Can the visual type change as data changes in the bookmark."""
    objects: BookmarkPartialVisualObject
    """Changes to formatting to apply in this bookmark."""
    orderBy: list[Orderby] | None = None
    """Updated ordering of data."""
    activeProjections: dict[str, list[Source]] | None = None
    """Updated projections that are used by the visual."""
    display: Display | None = None
    """Optional changes to how the visual is displayed."""
    expansionStates: list[ExpansionState] | None = None
    """Changes to expansion state."""
    targetType: str | None = None
    """Change visual to this type - if different from the original state. Used by personalize this visual on the web."""
    targetAutoSelectVisualType: bool = False
    """Change auto changing visual type."""
    projections: dict[str, list[Source]] | None = None
    """Projections are stored only when presentation changes are applied (such as during personalize visuals)
    or if the projections are resolved projections of a field parameter. When projections are resolved
    projections of a field parameter but presentation changes are not applied, nulls will be present in the
    array to indicate where non-parameter projections of the visual are."""
    parameters: dict[str, list[Parameter]] | None = None
    """Field parameters that were used as part of the query. We always capture parameter state when parameters
    are present regardless of the bookmark type."""
    cachedFilterDisplayItems: list[CachedDisplayNames] | None = None
    isDrillDisabled: bool = False
    """Indicates whether the drill feature is disabled"""


class BookmarkVisual(LayoutNode):
    filters: BookmarkFilters | None = None
    """State of filters for this visual when the bookmark was captured."""
    singleVisual: BookmarkPartialVisual
    """State of the configuration of the visual."""
    highlight: Highlight | None = None
    """"Cross-highlights captured in the bookmark."""


class VisualContainerGroup(LayoutNode):
    isHidden: bool = False
    children: dict[str, "VisualContainerGroup"] | None = None


class BookmarkSection(LayoutNode):
    _parent: "ExplorationState"  # pyright: ignore reportIncompatibleVariableOverride=false

    visualContainers: dict[str, BookmarkVisual]
    """Flat list of visual-container-specific state.
    Does not include state of groups."""
    filters: BookmarkFilters | None = None
    """State of filters for this page when the bookmark was captured."""
    visualContainerGroups: dict[str, VisualContainerGroup] | None = None
    """Flat list of group-specific state.\nDoes not include state of visual containers."""


class OutspacePaneProperties(LayoutNode):
    expanded: Expression | None = None
    visible: Expression | None = None


class OutspacePane(LayoutNode):
    properties: OutspacePaneProperties


class MergeProperties(LayoutNode):
    outspacePane: list[OutspacePane]


class ExplorationStateProperties(LayoutNode):
    merge: MergeProperties | None = None


class ExplorationState(LayoutNode):
    _parent: "Bookmark"  # pyright: ignore reportIncompatibleVariableOverride=false

    version: str
    """Version of bookmark."""
    sections: dict[str, BookmarkSection]
    """State of all pages when the bookmark was captured."""
    activeSection: str
    """Name of the page that was active when this bookmark was captured."""
    filters: BookmarkFilters | None = None
    """State of filters for the report when the bookmark was captured."""
    objects: ExplorationStateProperties | None = None
    """Changes to formatting to apply in this bookmark."""
    dataSourceVariables: str | None = None
    """A string containing the state of any variables from the underlying direct query data source that should be
    overridden when rendering this content.
    Data source variables do not supply values for M parameters in the semantic model. Instead, data source variables
    are applied when accessing the underlying direct query source."""


class BookmarkOptions(LayoutNode):
    _parent: "Bookmark"  # pyright: ignore reportIncompatibleVariableOverride=false

    applyOnlyToTargetVisuals: bool = False
    """Only applies changes to selected visuals when the bookmark was captured."""
    targetVisualNames: list[str] | None = None
    """Specific visuals to which this bookmark applies."""
    suppressActiveSection: bool = False
    """Don't apply changes to active section."""
    suppressData: bool = False
    """Don't apply data changes."""
    suppressDisplay: bool = False
    """Don't apply display property changes."""


class Bookmark(LayoutNode):
    """Defines a bookmark that captures the state of a report.

    Based on schema/bookmark.json
    """

    _parent: "Layout"  # pyright: ignore reportIncompatibleVariableOverride=false

    options: BookmarkOptions | None
    """Additional options for the bookmark."""
    explorationState: ExplorationState
    """Bookmark definition to use when applying this bookmark."""
    name: str
    """Unique identifier for the bookmark - unique across the report."""
    displayName: str
    """Display name for the bookmark."""

    def match_current_filters(self) -> None:
        raise NotImplementedError

    @staticmethod
    def new(  # noqa: PLR0913
        section: "Section",
        selected_visuals: list["BaseVisual"],
        bookmark_name: str,
        *,
        include_data: bool = True,
        include_display: bool = True,
        include_current_page: bool = True,
    ) -> "Bookmark":
        raise NotImplementedError


class BookmarkFolder(LayoutNode):
    _parent: "Layout"  # pyright: ignore reportIncompatibleVariableOverride=false
    displayName: str
    name: str  # acts as an ID
    children: list[Bookmark]


def get_bookmark_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "explorationState" in v:
            return "Bookmark"
        return "BookmarkFolder"
    return v.__class__.__name__


LayoutBookmarkChild = Annotated[
    Annotated[Bookmark, Tag("Bookmark")] | Annotated[BookmarkFolder, Tag("BookmarkFolder")],
    Discriminator(get_bookmark_type),
]



from pbi_core.static_files.layout._base_node import LayoutNode

from .base import SourceExpression


class ColumnSource(LayoutNode):
    Column: SourceExpression
    Name: str | None = None  # only seen on a couple TopN filters
    NativeReferenceName: str | None = None  # only for Layout.Visual.Query

    def __repr__(self) -> str:
        return f"ColumnSource({self.Column.Expression.table()}.{self.Column.Property})"

    def filter_name(self) -> str:
        return self.Column.Property

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ColumnSource):
            return False
        return (self.Column.column() == other.Column.column()) and (self.Column.table() == other.Column.table())

    def __hash__(self) -> int:
        return hash((self.Column.column(), self.Column.table()))



import datetime
from typing import TYPE_CHECKING, Literal
from uuid import UUID, uuid4

from pbi_parsers import dax
from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasRenameRecord, SsasTable
from pbi_core.ssas.model_tables.column import Column
from pbi_core.ssas.model_tables.enums import DataState, DataType
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.calc_dependency import CalcDependency
    from pbi_core.ssas.model_tables.detail_row_definition import DetailRowDefinition
    from pbi_core.ssas.model_tables.format_string_definition import FormatStringDefinition
    from pbi_core.ssas.model_tables.kpi import KPI
    from pbi_core.ssas.model_tables.table import Table
    from pbi_core.ssas.server.tabular_model.tabular_model import BaseTabularModel


class Measure(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/ab331e04-78f7-49f0-861f-3f155b8b4c5b)
    """

    data_category: str | None = None
    data_type: DataType
    description: str | None = None
    """This description will appear in the hover-over for the measure in the edit mode of the report"""

    detail_rows_definition_id: int | None = None
    display_folder: str | None = None
    error_message: str | None = None
    expression: str | int | float | None = None
    format_string: str | int | None = None
    """A static definition for the formatting of the measure"""

    format_string_definition_id: int | None = None
    """A foreign key to a FormatStringDefinition object. Used for dynamic-type measure formatting.

    Note:
        Mutually exclusive with format_string
    """
    is_hidden: bool = False
    """Controls whether the measure appears in the edit mode of the report"""
    is_simple_measure: bool
    kpi_id: int | None = None
    name: str
    """The name of the measure"""
    state: DataState
    table_id: int

    lineage_tag: UUID = uuid4()
    source_lineage_tag: UUID = uuid4()

    modified_time: datetime.datetime
    structure_modified_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.measure)

    def modification_hash(self) -> int:
        return hash((
            self.data_category,
            self.data_type,
            self.description,
            self.detail_rows_definition_id,
            self.display_folder,
            self.error_message,
            self.expression,
            self.format_string,
            self.format_string_definition_id,
            self.is_hidden,
            self.is_simple_measure,
            self.kpi_id,
            self.name,
            self.state,
            self.table_id,
            self.lineage_tag,
            self.source_lineage_tag,
            self.modified_time,
            self.structure_modified_time,
        ))

    def expression_ast(self) -> dax.Expression | None:
        if not isinstance(self.expression, str):
            return None
        ret = dax.to_ast(self.expression)
        if ret is None:
            msg = "Failed to parse DAX expression from measure"
            raise ValueError(msg)
        return ret

    def detail_rows_definition(self) -> "DetailRowDefinition | None":
        if self.detail_rows_definition_id is None:
            return None
        return self.tabular_model.detail_row_definitions.find(self.detail_rows_definition_id)

    def format_string_definition(self) -> "FormatStringDefinition | None":
        if self.format_string_definition_id is None:
            return None
        return self.tabular_model.format_string_definitions.find(self.format_string_definition_id)

    def KPI(self) -> "KPI | None":  # noqa: N802
        if self.kpi_id is not None:
            return self.tabular_model.kpis.find({"id": self.kpi_id})
        return None

    def table(self) -> "Table":
        """The Table object the measure is saved under."""
        return self.tabular_model.tables.find({"id": self.table_id})

    def full_name(self) -> str:
        """Returns the fully qualified name for DAX queries."""
        table_name = self.table().name
        return f"'{table_name}'[{self.name}]"

    def data(self, columns: Column | list[Column], head: int = 100) -> list[dict[str, int | float | str]]:
        if isinstance(columns, Column):
            columns = [columns]
        column_str = "\n".join(
            col.full_name() + "," for col in columns
        )  # this should eventually be converted to jinja imo
        command = f"""
        EVALUATE TOPN({head}, SUMMARIZECOLUMNS(
            {column_str}
            {columns[0].table().name},
            "measure", {self.full_name()}
        ))
        """
        return self.tabular_model.server.query_dax(command)

    def __repr__(self) -> str:
        return f"Measure({self.table().name}.{self.name})"

    def child_measures(self, *, recursive: bool = False) -> set["Measure"]:
        dependent_measures = self.tabular_model.calc_dependencies.find_all({
            "referenced_object_type": "MEASURE",
            "referenced_table": self.table().name,
            "referenced_object": self.name,
            "object_type": "MEASURE",
        })
        child_keys: list[tuple[str | None, str]] = [(m.table, m.object) for m in dependent_measures]
        full_dependencies = [m for m in self.tabular_model.measures if (m.table().name, m.name) in child_keys]

        if recursive:
            recursive_dependencies: set[Measure] = set()
            for dep in full_dependencies:
                if f"[{self.name}]" in str(dep.expression):
                    recursive_dependencies.add(dep)
                    recursive_dependencies.update(dep.child_measures(recursive=True))
            return recursive_dependencies

        return {x for x in full_dependencies if f"[{self.name}]" in str(x.expression)}

    def parent_measures(self, *, recursive: bool = False) -> set["Measure"]:
        """Calculated columns can use Measures too :(."""
        dependent_measures: set[CalcDependency] = self.tabular_model.calc_dependencies.find_all({
            "object_type": "MEASURE",
            "table": self.table().name,
            "object": self.name,
            "referenced_object_type": "MEASURE",
        })
        parent_keys = [(m.referenced_table, m.referenced_object) for m in dependent_measures]
        full_dependencies = [m for m in self.tabular_model.measures if (m.table().name, m.name) in parent_keys]

        if recursive:
            recursive_dependencies: set[Measure] = set()
            for dep in full_dependencies:
                if f"[{dep.name}]" in str(self.expression):
                    recursive_dependencies.add(dep)
                    recursive_dependencies.update(dep.parent_measures(recursive=True))
            return recursive_dependencies

        return {x for x in full_dependencies if f"[{x.name}]" in str(self.expression)}

    def child_columns(self, *, recursive: bool = False) -> set["Column"]:
        """Only occurs when the dependent column is calculated (expression is not None)."""
        dependent_measures = self.tabular_model.calc_dependencies.find_all({
            "referenced_object_type": "MEASURE",
            "referenced_table": self.table().name,
            "referenced_object": self.name,
        })
        child_keys = [(m.table, m.object) for m in dependent_measures if m.object_type in {"CALC_COLUMN", "COLUMN"}]
        full_dependencies = [m for m in self.tabular_model.columns if (m.table().name, m.explicit_name) in child_keys]

        if recursive:
            recursive_dependencies: set[Column] = set()
            for dep in full_dependencies:
                if f"[{self.name}]" in str(dep.expression):
                    recursive_dependencies.add(dep)
                    recursive_dependencies.update(dep.child_columns(recursive=True))
            return recursive_dependencies

        return {x for x in full_dependencies if f"[{self.name}]" in str(x.expression)}

    def parent_columns(self, *, recursive: bool = False) -> set["Column"]:
        """Only occurs when column is calculated."""
        dependent_measures = self.tabular_model.calc_dependencies.find_all({
            "object_type": "MEASURE",
            "table": self.table().name,
            "object": self.name,
        })
        parent_keys = {
            (m.referenced_table, m.referenced_object)
            for m in dependent_measures
            if m.referenced_object_type in {"CALC_COLUMN", "COLUMN"}
        }
        full_dependencies = [c for c in self.tabular_model.columns if (c.table().name, c.explicit_name) in parent_keys]

        if recursive:
            recursive_dependencies: set[Column] = set()
            for dep in full_dependencies:
                if f"[{dep.explicit_name}]" in str(self.expression):
                    recursive_dependencies.add(dep)
                    recursive_dependencies.update(dep.parent_columns(recursive=True))
            return recursive_dependencies

        return {x for x in full_dependencies if f"[{x.explicit_name}]" in str(self.expression)}

    def parents(self, *, recursive: bool = False) -> "set[Column | Measure]":
        """Returns all columns and measures this Measure is dependent on."""
        full_dependencies = self.parent_columns() | self.parent_measures()
        if recursive:
            recursive_dependencies: set[Column | Measure] = set()
            for dep in full_dependencies:
                recursive_dependencies.add(dep)
                recursive_dependencies.update(dep.parents(recursive=True))
            return recursive_dependencies

        return full_dependencies

    def children(self, *, recursive: bool = False) -> "set[Column | Measure]":
        """Returns all columns and measures dependent on this Measure."""
        full_dependencies = self.child_columns() | self.child_measures()
        if recursive:
            recursive_dependencies: set[Column | Measure] = set()
            for dep in full_dependencies:
                recursive_dependencies.add(dep)
                recursive_dependencies.update(dep.children(recursive=True))
            return recursive_dependencies

        return full_dependencies

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(
                self,
                lineage_type,
                [c.get_lineage(lineage_type) for c in self.child_measures() | self.child_columns()],
            )
        parent_nodes: list[SsasTable | None] = [
            self.KPI(),
            self.table(),
            *self.parent_measures(),
            *self.parent_columns(),
        ]
        parent_lineage = [c.get_lineage(lineage_type) for c in parent_nodes if c is not None]
        return LineageNode(self, lineage_type, parent_lineage)

    @staticmethod
    def new(
        name: str,
        expression: str,
        table: "int | Table",
        ssas: "BaseTabularModel",
    ) -> "Measure":
        table_id = table if isinstance(table, int) else table.id

        return Measure(
            name=name,
            data_type=DataType.UNKNOWN,
            is_simple_measure=True,
            state=DataState.READY,
            modified_time=datetime.datetime.now(tz=datetime.UTC),
            structure_modified_time=datetime.datetime.now(tz=datetime.UTC),
            expression=expression,
            table_id=table_id,
            tabular_model=ssas,
            id=-1,
        )



from pbi_core.static_files.layout._base_node import LayoutNode

from .base import Expression


# TODO: subclass filters so that the properties have fewer None defaults
class FilterProperties(LayoutNode):
    isInvertedSelectionMode: Expression | None = None
    requireSingleSelect: Expression | None = None


class FilterPropertiesContainer(LayoutNode):
    properties: FilterProperties


class FilterObjects(LayoutNode):
    general: list[FilterPropertiesContainer] | None = None



import datetime
import json
import math
import textwrap
import threading
import time
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import jinja2
from pbi_pyadomd import Connection, Reader

from pbi_core.logging import get_logger
from pbi_core.ssas.server.tabular_model.tabular_model import BaseTabularModel
from pbi_core.ssas.server.trace.trace_enums import TraceEvents

logger = get_logger()
TRACE_DIR = Path(__file__).parent / "templates"
TRACE_TEMPLATES: dict[str, jinja2.Template] = {
    f.name: jinja2.Template(f.read_text()) for f in TRACE_DIR.iterdir() if f.is_file()
}

TRACE_DEFAULT_EVENTS = (
    TraceEvents.DISCOVER_BEGIN,
    TraceEvents.COMMAND_BEGIN,
    TraceEvents.QUERY_END,
    TraceEvents.QUERY_SUBCUBE,
    TraceEvents.VERTIPAQ_SE_QUERY_BEGIN,
    TraceEvents.VERTIPAQ_SE_QUERY_END,
    TraceEvents.VERTIPAQ_SE_QUERY_CACHE_MATCH,
    TraceEvents.VERTIPAQ_SE_QUERY_CACHE_MISS,
    TraceEvents.AGGREGATE_TABLE_REWRITE_QUERY,
    TraceEvents.DIRECT_QUERY_END,
    TraceEvents.QUERY_BEGIN,
    TraceEvents.EXECUTION_METRICS,
)


def pretty_size(n: int) -> str:
    """Convert a number of bytes into a human-readable format."""
    units = ["B"] + [f"{p}iB" for p in "KMGTPEZY"]

    prefix = int(math.log(max(n, 1), 1024))
    prefix = min(prefix, len(units) - 1)
    pretty_n = round(n / 1024**prefix, 1)
    return f"{pretty_n} {units[prefix]}"


@dataclass
class ThreadResult:
    command: str
    rows_returned: int
    error: Exception | None = None

    def get_performance(self, trace_records: list[dict[str, Any]]) -> "Performance":
        event_info = {}
        for record in trace_records:
            event_info[record["EventClass"]] = record

        execution_metrics = json.loads(event_info["EXECUTION_METRICS"]["TextData"])
        command_text = event_info["QUERY_END"]["TextData"]
        start_datetime = datetime.datetime.strptime(execution_metrics["timeStart"], "%Y-%m-%dT%H:%M:%S.%f%z")
        end_datetime = datetime.datetime.strptime(execution_metrics["timeEnd"], "%Y-%m-%dT%H:%M:%S.%f%z")
        total_duration = execution_metrics["durationMs"]
        total_cpu_time = execution_metrics["totalCpuTimeMs"]
        query_cpu_time = execution_metrics["queryProcessingCpuTimeMs"]
        vertipaq_cpu_time = execution_metrics["vertipaqJobCpuTimeMs"]
        execution_delay = execution_metrics["executionDelayMs"]
        approximate_peak_consumption_kb = execution_metrics["approximatePeakMemConsumptionKB"]

        for record in trace_records:
            if record.get("EventClass") == "QUERY_END":
                return Performance(
                    command_text=command_text,
                    start_datetime=start_datetime,
                    end_datetime=end_datetime,
                    total_duration=total_duration,
                    total_cpu_time=total_cpu_time,
                    query_cpu_time=query_cpu_time,
                    vertipaq_cpu_time=vertipaq_cpu_time,
                    execution_delay=execution_delay,
                    approximate_peak_consumption_kb=approximate_peak_consumption_kb,
                    rows_returned=self.rows_returned,
                    full_trace_log=trace_records,
                )
        msg = f"Command '{self.command}' not found in trace records."
        raise ValueError(msg)


@dataclass
class Performance:
    command_text: str
    start_datetime: datetime.datetime
    end_datetime: datetime.datetime
    query_cpu_time: int
    # duration is the total time in milliseconds
    total_duration: int
    total_cpu_time: int
    vertipaq_cpu_time: int
    execution_delay: int
    approximate_peak_consumption_kb: int
    rows_returned: int
    full_trace_log: list[dict[str, Any]]

    def __repr__(self) -> str:
        total_duration = round(self.total_duration / 1000, 2)
        total_cpu_time = round(self.total_cpu_time / 1000, 2)
        approximate_peak_consumption = pretty_size(self.approximate_peak_consumption_kb * 1024)
        return f"Performance(rows={self.rows_returned}, total_duration={total_duration}, total_cpu_time={total_cpu_time}, peak_consumption={approximate_peak_consumption}"  # noqa: E501

    def pprint(self) -> str:
        """Pretty print the performance metrics."""
        command_text = textwrap.indent(self.command_text, " " * 8)
        return f"""Performance(
    Command:

{command_text}

    Start Time: {self.start_datetime.isoformat()}
    End Time: {self.end_datetime.isoformat()}
    Total Duration: {self.total_duration} ms
    Total CPU Time: {self.total_cpu_time} ms
    Query CPU Time: {self.query_cpu_time} ms
    Vertipaq CPU Time: {self.vertipaq_cpu_time} ms
    Execution Delay: {self.execution_delay} ms
    Approximate Peak Consumption: {pretty_size(self.approximate_peak_consumption_kb * 1024)}
    Rows Returned: {self.rows_returned}
)"""


class Subscriber:
    trace_records: dict[str, list[dict[str, Any]]]
    reader: Reader

    def __init__(self, subscription_create_command: str, conn: Connection, events: Iterable[TraceEvents]) -> None:
        self.reader = conn.execute_dax(
            subscription_create_command,
        )  # occasionally seems to hang, but might actually be bad queries??
        self.events = events
        self.trace_records = {}
        self.command_request_ids: dict[str, str] = {}
        self.thread = threading.Thread(target=self.poll_cursor, daemon=True)
        self.thread.start()

    def poll_cursor(self) -> None:
        event_mapping = {e.value: e.name for e in self.events}
        for record in self.reader.fetch_stream():
            if "EventClass" in record:
                record["EventClass"] = event_mapping[record["EventClass"]]
            self.trace_records.setdefault(record["RequestID"], []).append(record)
            if "TextData" in record:
                self.command_request_ids[record["TextData"]] = record["RequestID"]

    def kill_polling(self) -> None:
        if self.thread.is_alive():
            self.thread.join(timeout=1)

    def __repr__(self) -> str:
        return f"Subscriber(records={len(self.trace_records)})"


class PerformanceTrace:
    events: Iterable[TraceEvents]
    db: BaseTabularModel
    commands: list[str]
    trace_create_command: str
    subscription_create_command: str
    trace_delete_command: str
    clear_cache_command: str

    def __init__(
        self,
        db: BaseTabularModel,
        commands: list[str],
        events: Iterable[TraceEvents] = TRACE_DEFAULT_EVENTS,
    ) -> None:
        self.events = events
        self.db: BaseTabularModel = db
        self.commands = [x.replace("\r\n", "\n") for x in commands]

        next_day = (datetime.datetime.now(datetime.UTC) + datetime.timedelta(days=1)).strftime("%Y-%m-%dT%H:%M:%SZ")
        trace_name = f"pbi_core_{next_day.replace(':', '_')}"
        subscription_name = f"pbi_core_subscription_{next_day.replace(':', '_')}"
        self.trace_create_command = TRACE_TEMPLATES["trace_create.xml"].render(
            trace_name=trace_name,
            stop_time=next_day,
            events=self.events,
        )
        self.subscription_create_command = TRACE_TEMPLATES["subscription_create.xml"].render(
            trace_name=trace_name,
            subscription_name=subscription_name,
        )
        self.trace_delete_command = TRACE_TEMPLATES["trace_delete.xml"].render(
            trace_name=trace_name,
        )
        self.clear_cache_command = TRACE_TEMPLATES["clear_cache.xml"].render(database_name=self.db.db_name)

    def initialize_tracing(self) -> "PerformanceTrace":
        logger.info("Beginning trace")
        self.db.server.query_xml(self.trace_create_command)
        self.subscriber = Subscriber(
            self.subscription_create_command,
            self.get_conn(),
            self.events,
        )
        return self

    def terminate_tracing(self) -> None:
        logger.info("Terminating trace")
        with self.db.server.conn(db_name=self.db.db_name) as conn:
            conn.execute_xml(self.trace_delete_command)

    def get_performance(self, *, clear_cache: bool = False) -> list[Performance]:
        def thread_func(command: str) -> ThreadResult:
            try:
                with self.get_conn() as conn:
                    reader = conn.execute_dax(command)
                    rows_returned = len(reader.fetch_many(limit=500))

                    # See note in Reader.fetch_stream() documentation for why we do this
                    reader.close()

            except ValueError as e:
                return ThreadResult(
                    command=command,
                    rows_returned=-1,
                    error=e,
                )
            # The limit of 500 is to mimic the behavior of PowerBI, which returns by defult 500 rows
            return ThreadResult(
                command=command,
                rows_returned=rows_returned,
            )

        self.initialize_tracing()
        if clear_cache:
            logger.info("Clearing cache before performance testing")
            with self.db.server.conn(db_name=self.db.db_name) as conn:
                conn.execute_xml(self.clear_cache_command)

        with ThreadPoolExecutor() as dax_executor:
            logger.info("Running DAX commands")
            command_results = list(dax_executor.map(thread_func, self.commands))

        missing_subscription_records = command_results
        for _ in range(5):
            # Testing for 15 seconds to check that all commands have an entry in the trace
            new_missing_subscription_records = []
            for command in missing_subscription_records:
                request_id = self.subscriber.command_request_ids.get(command.command)
                if request_id is None:
                    new_missing_subscription_records.append(command)
                    continue
                command_events = self.subscriber.trace_records.get(request_id, [])
                if not any(x["EventClass"] == "QUERY_END" for x in command_events):
                    new_missing_subscription_records.append(command)

            if not new_missing_subscription_records:
                break
            time.sleep(1)
            missing_subscription_records = new_missing_subscription_records
        else:
            missing_commands = ", ".join(cmd.command for cmd in missing_subscription_records)
            msg = f"Some commands did not have trace records: {missing_commands}"
            raise ValueError(msg)

        self.subscriber.kill_polling()
        self.terminate_tracing()

        return [
            command_result.get_performance(
                self.subscriber.trace_records[self.subscriber.command_request_ids[command_result.command]],
            )
            for command_result in command_results
        ]

    def get_conn(self) -> Connection:
        return self.db.server.conn(db_name=self.db.db_name).open()



import json
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import openpyxl

from pbi_core import LocalReport
from pbi_core.static_files import Layout, Section

if TYPE_CHECKING:
    from _typeshed import StrPath


@dataclass
class StaticElement:
    xpath: list[str | int]
    field: str
    text: str


class StaticElements:
    static_elements: dict[str, list[StaticElement]]

    def __init__(self) -> None:
        self.static_elements = {}

    def to_csv(self) -> None:
        pass

    def to_excel(self, path: "StrPath") -> None:
        wb = openpyxl.Workbook()
        for object_type, objects in self.static_elements.items():
            ws = wb.create_sheet(object_type)
            for j, name in enumerate(["xpath", "field", "default"]):
                ws.cell(1, j + 1).value = name
            for i, obj in enumerate(objects):
                ws.cell(2 + i, 1).value = json.dumps(obj.xpath)
                ws.cell(2 + i, 2).value = obj.field
                ws.cell(2 + i, 3).value = obj.text
        wb.remove(wb["Sheet"])
        wb.save(path)


def get_static_elements(layout: Layout) -> StaticElements:
    ret = StaticElements()
    for section in layout.find_all(Section):
        ret.static_elements.setdefault("section", []).append(
            StaticElement(
                xpath=section._xpath,
                field="displayName",
                text=section.displayName,
            ),
        )
    return ret


def set_static_elements(translation_path: "StrPath", pbix_path: "StrPath") -> None:
    wb = openpyxl.load_workbook(translation_path)
    languages: list[str] = [str(x) for x in next(iter(wb.worksheets[0].values))[3:]]
    processing: dict[str, list[StaticElement]] = {}
    for ws in wb.worksheets:
        for row in list(ws.values)[1:]:
            for i, language in enumerate(languages):
                processing.setdefault(language, []).append(
                    StaticElement(json.loads(str(row[0])), str(row[1]), str(row[3 + i])),
                )
    for language, static_elements in processing.items():
        pbix = LocalReport.load_pbix(pbix_path)
        for static_element in static_elements:
            node = pbix.static_files.layout.find_xpath(static_element.xpath)
            setattr(node, static_element.field, static_element.text)
        out_path = f"{Path(pbix_path).with_suffix('').absolute().as_posix()}_{language}.pbix"
        pbix.save_pbix(out_path)



from enum import IntEnum


class MetadataPermission(IntEnum):
    """Source: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/ac2ceeb3-a54e-4bf5-85b0-a770d4b1716e)."""

    DEFAULT = 0
    _NONE = 1
    READ = 2



from enum import IntEnum


class VertipaqSeQueryBeginColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    STARTTIME = 3
    JOBID = 7
    SESSIONTYPE = 8
    OBJECTID = 11
    OBJECTTYPE = 12
    OBJECTNAME = 13
    OBJECTPATH = 14
    OBJECTREFERENCE = 15
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    NTDOMAINNAME = 33
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERNAME = 43
    ACTIVITYID = 46
    REQUESTID = 47
    USEROBJECTID = 51
    APPLICATIONCONTEXT = 52
    DATABASEFRIENDLYNAME = 54
    IDENTITY = 55


class VertipaqSeQueryEndColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    JOBID = 7
    SESSIONTYPE = 8
    PROGRESSTOTAL = 9
    INTEGERDATA = 10
    OBJECTID = 11
    OBJECTTYPE = 12
    OBJECTNAME = 13
    OBJECTPATH = 14
    OBJECTREFERENCE = 15
    SEVERITY = 22
    SUCCESS = 23
    ERROR = 24
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    NTDOMAINNAME = 33
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERNAME = 43
    ACTIVITYID = 46
    REQUESTID = 47
    ERRORTYPE = 49
    USEROBJECTID = 51
    APPLICATIONCONTEXT = 52
    DATABASEFRIENDLYNAME = 54
    IDENTITY = 55


class VertipaqSeQueryCacheMissColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    JOBID = 7
    SESSIONTYPE = 8
    OBJECTID = 11
    OBJECTTYPE = 12
    OBJECTNAME = 13
    OBJECTPATH = 14
    OBJECTREFERENCE = 15
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    NTDOMAINNAME = 33
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    USEROBJECTID = 51
    APPLICATIONCONTEXT = 52
    DATABASEFRIENDLYNAME = 54
    IDENTITY = 55


class VertipaqSeQueryCacheMatchColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    JOBID = 7
    SESSIONTYPE = 8
    OBJECTID = 11
    OBJECTTYPE = 12
    OBJECTNAME = 13
    OBJECTPATH = 14
    OBJECTREFERENCE = 15
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    NTDOMAINNAME = 33
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    USEROBJECTID = 51
    APPLICATIONCONTEXT = 52
    DATABASEFRIENDLYNAME = 54
    IDENTITY = 55


class CalculationEvaluationColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    INTEGERDATA = 10
    CONNECTIONID = 25
    DATABASENAME = 28
    CLIENTHOSTNAME = 35
    CLIENTPROCESSID = 36
    APPLICATIONNAME = 37
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    APPLICATIONCONTEXT = 52


class CalculationEvaluationDetailedInformationColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    INTEGERDATA = 10
    CONNECTIONID = 25
    DATABASENAME = 28
    CLIENTHOSTNAME = 35
    CLIENTPROCESSID = 36
    APPLICATIONNAME = 37
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    APPLICATIONCONTEXT = 52


class DaxQueryPlanColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    INTEGERDATA = 10
    CONNECTIONID = 25
    DATABASENAME = 28
    CLIENTHOSTNAME = 35
    CLIENTPROCESSID = 36
    APPLICATIONNAME = 37
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    APPLICATIONCONTEXT = 52


class DaxExtensionExecutionBeginColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    INTEGERDATA = 10
    DATABASENAME = 28
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    FUNCTIONNAME = 50
    APPLICATIONCONTEXT = 52


class DaxExtensionExecutionEndColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    INTEGERDATA = 10
    SEVERITY = 22
    SUCCESS = 23
    ERROR = 24
    DATABASENAME = 28
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    ERRORTYPE = 49
    FUNCTIONNAME = 50
    APPLICATIONCONTEXT = 52


class DaxExtensionTraceErrorColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    INTEGERDATA = 10
    DATABASENAME = 28
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    FUNCTIONNAME = 50
    APPLICATIONCONTEXT = 52


class DaxExtensionTraceInfoColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    INTEGERDATA = 10
    DATABASENAME = 28
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    FUNCTIONNAME = 50
    APPLICATIONCONTEXT = 52


class DaxExtensionTraceVerboseColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    INTEGERDATA = 10
    DATABASENAME = 28
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    FUNCTIONNAME = 50
    APPLICATIONCONTEXT = 52


class AggregateTableRewriteQueryColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    INTEGERDATA = 10
    SUCCESS = 23
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    NTDOMAINNAME = 33
    CLIENTHOSTNAME = 35
    CLIENTPROCESSID = 36
    APPLICATIONNAME = 37
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    USEROBJECTID = 51
    APPLICATIONCONTEXT = 52


class AggregateTableRewriteInfoColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    STARTTIME = 3
    SEVERITY = 22
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    NTDOMAINNAME = 33
    CLIENTHOSTNAME = 35
    CLIENTPROCESSID = 36
    APPLICATIONNAME = 37
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    APPLICATIONCONTEXT = 52


class QuerySubcubeColumns(IntEnum):
    EVENTCLASS = 0
    EVENTSUBCLASS = 1
    CURRENTTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    OBJECTPATH = 14
    CONNECTIONID = 25
    DATABASENAME = 28
    SESSIONID = 39
    NTCANONICALUSERNAME = 40
    SPID = 41
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    APPLICATIONCONTEXT = 52
    DATABASEFRIENDLYNAME = 54
    IDENTITY = 55


class DirectQueryBeginColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    JOBID = 7
    SESSIONTYPE = 8
    INTEGERDATA = 10
    OBJECTID = 11
    OBJECTTYPE = 12
    OBJECTNAME = 13
    OBJECTPATH = 14
    SEVERITY = 22
    SUCCESS = 23
    ERROR = 24
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    CLIENTPROCESSID = 36
    SESSIONID = 39
    SPID = 41
    TEXTDATA = 42
    SERVERNAME = 43
    ACTIVITYID = 46
    REQUESTID = 47
    ERRORTYPE = 49
    APPLICATIONCONTEXT = 52
    DATABASEFRIENDLYNAME = 54
    IDENTITY = 55


class DirectQueryEndColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    JOBID = 7
    SESSIONTYPE = 8
    INTEGERDATA = 10
    OBJECTID = 11
    OBJECTTYPE = 12
    OBJECTNAME = 13
    OBJECTPATH = 14
    SEVERITY = 22
    SUCCESS = 23
    ERROR = 24
    CONNECTIONID = 25
    DATABASENAME = 28
    NTUSERNAME = 32
    CLIENTPROCESSID = 36
    SESSIONID = 39
    SPID = 41
    TEXTDATA = 42
    SERVERNAME = 43
    ACTIVITYID = 46
    REQUESTID = 47
    ERRORTYPE = 49
    APPLICATIONCONTEXT = 52
    DATABASEFRIENDLYNAME = 54
    IDENTITY = 55


class ExecutionMetricsColumns(IntEnum):
    EVENTCLASS = 0
    DATABASENAME = 28
    APPLICATIONNAME = 37
    SPID = 41
    TEXTDATA = 42
    ACTIVITYID = 46
    REQUESTID = 47



from collections.abc import Callable
from typing import Any, Final, TypeVar

import pydantic


class IdBase:
    id: Final[int] = pydantic.Field(frozen=True)


T = TypeVar("T", bound=IdBase)


class RowNotFoundError(Exception):
    pass


class Group(list[T]):  # noqa: FURB189
    def find(self, match_val: int | dict[str, Any] | Callable[[T], bool]) -> T:
        """Gets a matching value from the group.

        Finds a matching SSAS record from the group using the following rules:
        1. If match_val is an int, it finds a record with a matching `id`
        2. If match_val is a dictionary, it uses the keys as field names and values as field values.
            It returns the first record to match all items
        3. If match_val is a function, it returns the first record to return true when passed to the function

        Raises:
            RowNotFoundError: when no value matches the match_val

        """
        if isinstance(match_val, int):
            for val in self:
                if val.id == match_val:
                    return val
        elif isinstance(match_val, dict):
            for val in self:
                if all(getattr(val, field_name) == field_value for field_name, field_value in match_val.items()):
                    return val

        else:
            for val in self:
                if match_val(val) is True:
                    return val
        raise RowNotFoundError

    def find_all(self, match_val: int | dict[str, Any] | Callable[[T], bool]) -> set[T]:
        ret: set[T] = set()
        if isinstance(match_val, int):
            ret.update(val for val in self if val.id == match_val)
        elif isinstance(match_val, dict):
            ret.update(
                val
                for val in self
                if all(getattr(val, field_name) == field_value for field_name, field_value in match_val.items())
            )
        else:
            ret.update(val for val in self if match_val(val) is True)
        return ret

    def sync_to_server(self) -> None:
        pass

    def sync_from_server(self) -> None:
        pass



import datetime
from typing import TYPE_CHECKING, Literal

from bs4 import BeautifulSoup
from pbi_parsers import dax, pq
from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.logging import get_logger
from pbi_core.ssas.model_tables._group import RowNotFoundError
from pbi_core.ssas.model_tables.base import RefreshType, SsasRefreshRecord, SsasTable
from pbi_core.ssas.model_tables.column import Column
from pbi_core.ssas.model_tables.enums import DataState
from pbi_core.ssas.server._commands import RefreshCommands
from pbi_core.ssas.server.utils import SsasCommands

from .enums import DataView, PartitionMode, PartitionType

if TYPE_CHECKING:
    from collections.abc import Iterable

    from pbi_core.ssas.model_tables.expression import Expression
    from pbi_core.ssas.model_tables.query_group import QueryGroup
    from pbi_core.ssas.model_tables.table import Table


logger = get_logger()


class Partition(SsasRefreshRecord):
    """Partitions are a child of Tables. They contain the Power Query code.

    These are the physical segments of the table that contain the data.
    They cannot be edited within the Power BI Desktop UI, but can be edited in the
    Tabular Editor or other tools (like this one!). Data refreshes occur on the Partition-level.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/81badb81-31a8-482b-ae16-5fc9d8291d9e)
    """

    _default_refresh_type = RefreshType.FULL

    data_view: DataView
    data_source_id: int | None = None
    description: str | None = None
    error_message: str | None = None
    expression_source_id: int | None = None
    m_attributes: str | None = None
    mode: PartitionMode
    name: str
    partition_storage_id: int
    query_definition: str
    query_group_id: int | None = None
    range_granularity: int
    retain_data_till_force_calculate: bool
    state: DataState
    system_flags: int
    table_id: int
    type: PartitionType

    modified_time: datetime.datetime
    refreshed_time: datetime.datetime

    _commands: RefreshCommands = PrivateAttr(default_factory=lambda: SsasCommands.partition)

    def expression_ast(self) -> dax.Expression | pq.Expression | None:
        if self.type == PartitionType.CALCULATED:
            ret = pq.to_ast(self.query_definition)
            if ret is None:
                msg = "Failed to parse DAX expression from partition query definition"
                raise ValueError(msg)
        elif self.type == PartitionType.M:
            ret = dax.to_ast(self.query_definition)
            if ret is None:
                msg = "Failed to parse M expression from partition query definition"
                raise ValueError(msg)
        else:
            logger.warning("Attempted to get AST of non-M/DAX partition", partition=self.name, type=self.type)
            return None

    def expression_source(self) -> "Expression | None":
        if self.expression_source_id is None:
            return None
        return self.tabular_model.expressions.find(self.expression_source_id)

    def is_system_table(self) -> bool:
        return bool(self.system_flags >> 1 % 2)

    def is_from_calculated_table(self) -> bool:
        return bool(self.system_flags % 2)

    def query_group(self) -> "QueryGroup | None":
        try:
            return self.tabular_model.query_groups.find({"id": self.table_id})
        except RowNotFoundError:
            return None

    def table(self) -> "Table":
        return self.tabular_model.tables.find({"id": self.table_id})

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)
        parent_nodes: list[SsasTable | None] = [self.table(), self.query_group()]
        parent_lineage: list[LineageNode] = [c.get_lineage(lineage_type) for c in parent_nodes if c is not None]
        return LineageNode(self, lineage_type, parent_lineage)

    def remove_columns(self, dropped_columns: "Iterable[Column | str | None]") -> BeautifulSoup:
        def pq_escape(x: str) -> str:
            """Beginning of column escaping for power query."""
            return x.replace('"', '""')

        """Adds a Table.RemoveColumns statement to the end of the Partition's PowerQuery.

        This means the upon refresh, the columns will not be included in the table
        """
        new_dropped_columns: list[str] = []
        for col in dropped_columns:
            if isinstance(col, Column):
                # Tables have a column named "RowNumber-<UUID>" that cannot be removed in the PowerQuery
                if col._column_type() != "CALC_COLUMN" and not col.is_key:
                    assert col.explicit_name is not None, "Column must have an explicit name to be dropped"
                    new_dropped_columns.append(col.explicit_name)
            elif isinstance(col, str):
                new_dropped_columns.append(col)

        # TODO: create a powerquery parser to do this robustly
        new_dropped_columns = [pq_escape(x) for x in new_dropped_columns]
        logger.info("Updating partition to drop columns", table=self.table().name, columns=new_dropped_columns)
        lines = self.query_definition.split("\n")
        final_table_name = lines[-1].strip()
        setup = "\n".join(lines[:-2])

        prior_updates = setup.count("pbi_update")  # used to keep statement variables unique when applied multiple times
        new_final_table_name = f"pbi_update{prior_updates}"

        cols = ", ".join(f'"{x}"' for x in new_dropped_columns)
        setup += f",\n    {new_final_table_name} = Table.RemoveColumns({final_table_name}, {{{cols}}})"
        setup += f"\nin\n    {new_final_table_name}"
        self.query_definition = setup
        return self.alter()



import datetime
from enum import IntEnum
from typing import TYPE_CHECKING, Literal
from uuid import UUID, uuid4

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasRenameRecord, SsasTable
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.column import Column
    from pbi_core.ssas.model_tables.model import Model
    from pbi_core.ssas.model_tables.query_group import QueryGroup
from pbi_core.lineage import LineageNode


class Kind(IntEnum):
    M = 0


class Expression(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/61f98e45-d5e3-4435-b829-2f2f043839c1)
    """

    description: str | None = None
    expression: str
    kind: Kind
    model_id: int
    name: str
    parameter_values_column_id: int | None = None
    query_group_id: int | None = None

    lineage_tag: UUID = uuid4()
    source_lineage_tag: UUID = uuid4()

    modified_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.expression)

    def model(self) -> "Model":
        return self.tabular_model.model

    def parameter_values_column(self) -> "Column | None":
        if self.parameter_values_column_id is None:
            return None
        return self.tabular_model.columns.find(self.parameter_values_column_id)

    def query_group(self) -> "QueryGroup | None":
        return self.tabular_model.query_groups.find({"id": self.query_group_id})

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)
        parent_nodes: list[SsasTable | None] = [self.model(), self.query_group()]
        parent_lineage = [p.get_lineage(lineage_type) for p in parent_nodes if p is not None]
        return LineageNode(self, lineage_type, parent_lineage)



import atexit
import pathlib
import shutil
import subprocess  # nosec. It's necessary to run the msmdsrv exe  # noqa: S404
from functools import cached_property
from typing import TYPE_CHECKING

import backoff
import psutil

from pbi_core.logging import get_logger
from pbi_core.ssas.setup import PbiCoreStartupConfig, get_startup_config

from .utils import get_msmdsrv_info

if TYPE_CHECKING:
    from _typeshed import StrPath
logger = get_logger()
PORT_ACCESS_TRIES = 5


class SSASProcess:
    """Handles the SSAS instance as a OS process.

    Args:
        pid (Optional[int]): The process ID of the SSAS instance. If None, the class will create a new SSAS process
        workspace_directory (StrPath): The path to the workspace corresponding to the pid.
            Should only be included when the pid is not provided.
        kill_on_exit (bool): Specifies whether the SSAS instance should be terminated when the python session exits
            (implemented via atexit lib)

    Examples:
        ```python

           SSASProcess('tmp/workspace')  # generates a new SSAS Process and generates workspace documents at 'tmp/workspace'
           SSASProcess(4321)  # connects to an existing SSAS Process at 4321
           SSASProcess('tmp/workspace', False)  # Generates a new SSAS Process and allows it to run beyond the lifetime of the Python script
        ```

    Raises:
        ValueError: If both or none of ``pid`` and ``workspace_directory`` are specified

    """  # noqa: E501

    _workspace_directory: pathlib.Path
    pid: int = -1
    kill_on_exit: bool
    startup_config: PbiCoreStartupConfig | None

    def __init__(
        self,
        pid: int | None = None,
        workspace_directory: "StrPath | None" = None,
        *,
        kill_on_exit: bool = True,
        startup_config: PbiCoreStartupConfig | None = None,
    ) -> None:
        """__init__ is not intended to be directly called.

        This class is expected to be initialized from the .from_pid or .new_instance methods

        Raises:
            ValueError: when either both or neither of the pid and workspace_directory are specified

        """
        self.kill_on_exit = kill_on_exit
        atexit.register(self._on_exit)

        if pid is None:
            if workspace_directory is None:
                msg = "If the pid is not specified, you must specify a workspace directory"
                raise ValueError(msg)
            self.startup_config = startup_config or get_startup_config()
            self._workspace_directory = pathlib.Path(workspace_directory)
            logger.info(
                "No pid provided. Initializing new SSAS Instance",
                workspace_dir=self._workspace_directory,
                msmdsrv_exe=self.startup_config.msmdsrv_exe,
            )
            self.pid = self._initialize_server()
        else:
            if workspace_directory is not None:
                msg = "If the pid is specified, you must not specify a workspace directory"
                raise ValueError(msg)
            self.startup_config = None
            self.pid = pid
            self._workspace_directory = self._get_workspace_directory()

    def _get_workspace_directory(self) -> pathlib.Path:
        """Gets the workspace of the PBIX DB config.

        Uses the PID and the command used to initialize that PID to identify the
        workspace directory used by the SSAS process

        Raises:
            ValueError: When the PID points to a non-DB process

        """
        proc = psutil.Process(self.pid)
        proc_info = get_msmdsrv_info(proc)
        if proc_info is None:
            msg = "This PID doesn't correspond to a valid SSAS instance"
            raise ValueError(msg)
        return proc_info.workspace_directory.absolute()

    def _create_workspace(self) -> None:
        """Creates the workspace directory and populates the initial config file for the new SSAS instance."""
        assert self.startup_config is not None, "Startup config must be set before creating a workspace"
        logger.debug("initializing SSAS Workspace", directory=self._workspace_directory)
        self._workspace_directory.mkdir(parents=True, exist_ok=True)
        (self._workspace_directory / "msmdsrv.ini").write_text(
            self.startup_config.msmdsrv_ini_template().render(
                data_directory=self._workspace_directory.as_posix().replace("/", "\\"),
                certificate_directory=self.startup_config.cert_dir.absolute().as_posix().replace("/", "\\"),
            ),
        )

    def _run_msmdsrv(self) -> int:
        """Runs the commands to create the DB.

        Commands are explained here: https://stackoverflow.com/q/36458981
        -c: console mode
        -n "instance_name": instance name (the name of the default database)
        -s "workspace_directory": The location of the configuration

        Note:
            ``-s`` points to the workspace created in the method "create_workspace"

        """
        assert self.startup_config is not None, "Startup config must be set before running msmdsrv"
        logger.debug("Running msmdsrv exe")
        command = [  # pbi_core_master is not really used, but a port file isn't generated without it
            self.startup_config.msmdsrv_exe.as_posix(),
            "-c",
            "-n",
            "pbi_core_master",
            "-s",
            self._workspace_directory.as_posix(),
        ]
        flags = subprocess.DETACHED_PROCESS | subprocess.CREATE_NEW_PROCESS_GROUP
        return subprocess.Popen(  # noqa: S603
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=flags,
        ).pid  # nosec. It's necessary to run the msmdsrv exe

    @cached_property
    @backoff.on_exception(backoff.expo, FileNotFoundError, max_time=5)
    def _port(self) -> int:
        """Gets the port of the DB.

        We include exponential backoff in this function since it occasionally takes 1-3 seconds for msmdsrv.exe
        to generate the msmdsrv.port.txt file in the workspace

        Raises:
            FileNotFoundError: when ``msmdsrv.port.txt`` cannot be found in the SSAS workspace folder

        """
        try:
            return int((self._workspace_directory / "msmdsrv.port.txt").read_text(encoding="utf-16-le"))
        except FileNotFoundError as e:
            msg = f"Could not find msmdsrv.port.txt file in directory {self._workspace_directory}. This is needed to get the port of the SSAS instance"  # noqa: E501
            raise FileNotFoundError(msg) from e

    def _initialize_server(self) -> int:
        self._create_workspace()
        return self._run_msmdsrv()

    def _on_exit(self) -> None:
        if self.kill_on_exit:
            logger.info("Terminating SSAS Process", pid=self.pid)
            self.terminate()

    @staticmethod
    @backoff.on_exception(backoff.expo, ValueError, max_time=10)
    def _wait_until_terminated(process: psutil.Process) -> None:
        """Takes a process class and checks if the process is still running.

        Raises:
            ValueError: When the process is running

        """
        if process.is_running():
            msg = "The process will not terminate"
            raise ValueError(msg)

    def terminate(self) -> None:
        """Kills the SSAS instance.

        The code performs the following:

        1. Checks the PID. If the PID isn't associated with an active process, we declare the SSAS instance killed
        2. Checks the information associated with the PID (from ``get_msmdsrv_info``).
            If it's not running ``msmdsrv.exe`` with active ports, we consider it killed
        3. We call a terminate command
        4. We wait until the command is in a non-running state
        5. We then remove the corresponding workspace
        """
        try:
            p = psutil.Process(self.pid)
        except psutil.NoSuchProcess:  # something else killed it??
            logger.info("SSAS Proc already terminated", pid=self.pid)
            return
        except ValueError:
            logger.info("SSAS Proc never initialized", pid=self.pid)
            return
        if not get_msmdsrv_info(p):  # indicates another process has already taken this PID
            return
        p.terminate()
        self._wait_until_terminated(p)
        logger.info("Terminated SSAS Proc", pid=self.pid)

        shutil.rmtree(self._workspace_directory, ignore_errors=True)
        logger.info("Workspace Removed", directory=self._workspace_directory.as_posix())



import datetime
from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.model_tables.enums import DataState
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

from .enums import MetadataPermission

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.role import Role


class TablePermission(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/ac2ceeb3-a54e-4bf5-85b0-a770d4b1716e)
    """

    error_message: str | None = None
    filter_expression: str | None = None
    metadata_permission: MetadataPermission
    role_id: int
    state: DataState
    table_id: int

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.table_permission)

    def role(self) -> "Role":
        return self.tabular_model.roles.find(self.role_id)



from typing import TYPE_CHECKING

from pbi_core.pydantic.main import BaseValidation

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.column import Column
    from pbi_core.ssas.model_tables.measure import Measure
    from pbi_core.ssas.model_tables.table import Table
    from pbi_core.ssas.server.tabular_model.tabular_model import BaseTabularModel


class ModelColumnReference(BaseValidation):
    column: str
    table: str

    def to_model(self, tabular_model: "BaseTabularModel") -> "Column":
        return tabular_model.columns.find(lambda c: (c.explicit_name == self.column) and (c.table().name == self.table))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModelColumnReference):
            return False
        return (self.column == other.column) and (self.table == other.table)

    def __hash__(self) -> int:
        return hash((self.column, self.table))


class ModelTableReference(BaseValidation):
    table: str

    def to_model(self, tabular_model: "BaseTabularModel") -> "Table":
        return tabular_model.tables.find(lambda t: (t.name == self.table))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModelTableReference):
            return False
        return self.table == other.table

    def __hash__(self) -> int:
        return hash(self.table)


class ModelMeasureReference(BaseValidation):
    measure: str
    table: str

    def to_model(self, tabular_model: "BaseTabularModel") -> "Measure":
        return tabular_model.measures.find(lambda m: (m.name == self.measure) and (m.table().name == self.table))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModelMeasureReference):
            return False
        return (self.measure == other.measure) and (self.table == other.table)

    def __hash__(self) -> int:
        return hash((self.measure, self.table))
